unit DelDirective;


{ ******************************************************** }
{                                                          }
{  DelDirective                                            }
{                                                          }
{  Copyright (c)2005, 2007 M.Braun                         }
{  Contact: info@grizzlymotion.com                         }
{                                                          }
{  Routines to remove comments and strings from Pascal     }
{  Source code                                             }
{  Includes a Pascal Preprocessor to remove code fragments }
{  that are outruled by compiler directives.               }
{                                                          }
{  History                                                 }
{  21.11.2006  Added extraction of routine list.           }
{  15.11.2006  Added extraction of comments.               }
{  03.09.2006  Added RTF export with syntax highlighting   }
{  06.07.2006  Added key word detection. Needed for        }
{              strings with " inside ASM statement         }
{  03.07.2006  Added more compiler directives and support  }
{              for "fat" switches                          }
{  17.06.2006  1st draft "Open source" edition             }
{              (and removed metrics part)                  }
{  23.01.2006  Re-implemented lost fix from 07.12.2005 8(  }
{  15.12.2005  Added simple metrics (slow!);               }
{              #0 allowed in source files                  }
{  07.12.2005  Bugfix (In (*$A+B-*) ignore "B-")           }
{  25.11.2005  Interface redesigned                        }
{  11.10.2005  Include files                               }
{  16.09.2005  Optional deleting of string-contents        }
{  15.04.2005  Handling of missing compiler settings       }
{  11.04.2005  Bug fixes                                   }
{  08.04.2005  Initial Version; M. Braun                   }
{                                                          }
{ ******************************************************** }


// ToDo:  "InsideImplementation"

interface




USES StrUtils, Classes, SysUtils;


{$ifdef Something}
  This should not be part of your sourcecode.
  'Hello';
{$endif}


CONST
  RTFDefLine1 = '{\rtf1\ansi\ansicpg1252\deff0\deflang1031{\fonttbl{\f0\fnil\fcharset0 Courier New;}}';
  RTFDefLine2 = '{\colortbl ;\red0\green0\blue0;\red0\green0\blue128;\red128\green128\blue160;\red32\green192\blue32;\red255\green255\blue128;\red192\green192\blue192;}\viewkind4\uc1\pard\cf1\f0\fs16';

VAR
  StrictDefCheck: boolean = false;  


{ ---------= Class definitions =---------------------------------------}
TYPE
  TParseAction   = set of (PA_None, PA_DelStrings, PA_DelComments, PA_InvDelComments, PA_DelDirectives, PA_BuildRTF, PA_ListRoutines);

  { For returning processed source files                                }
  { This class adds just some informative variables to a stringlist     }
  { Additionally it can read text files containing #0 which will simply }
  { be replayced by space (#32)                                         }
  TAnnotatedStringList = class(TStringlist)
                           public
                             SourceName: string;
                             CalledBy  : string;
                             Line      : integer;
                             procedure Free;
                             procedure Clear; override;
                             procedure Assign(Source: TPersistent); override;
                             procedure LoadFromStream(Stream: TStream); override;
                         end;

  { For returning a series of processed source files                    }
  { This class is a TList for a series of TAnnotatedStringList objects  }
  TSourceList          = class(TList)
                           public
                             procedure Add(ASL: TAnnotatedStringList);
                             procedure Clear; override;
                             procedure Free;
                         end;





{ ---------= The main functionality =----------------------------------}

FUNCTION RemoveDeadCodeEx(const SourceFileName: string;           // Name of source file to be processed. If empty SLOriginal must contain the source code already
                          const CalledBySource: string;           // "Calling" source file. Usually empty. Only of interest while processing include files
                                CalledFromLine: integer;          // Line numer of call. Usually undefined. Only of interest while processing include files
                                SLOriginal    : TSourceList;      // Series of TAnnotatedStringList objects. First (index 0) containes original source code.
                                                                  // The others contain original source code of include files.
                                SLProcessed   : TSourceList;      // Series of TAnnotatedStringList objects. First (index 0) containes processed source code.
                                                                  // The others contain processed source code of include files.
                                SLSourceRTF   : TSourceList;
                                SLRoutines    : TSourceList;
                                ParseAction   : TParseAction;     // "Todo" list
                          const SearchPath    : string;           // Searchpath to be used for include file location
                                DefList       : TStringList;      // String list containing all conditional defines and compiler settings. Will be updated during parsing
                            VAR ErrMsg        : string): integer; // Error message string.

FUNCTION RemoveDeadCode(const SourceFileName: string;             // Name of source file to be processed. If empty SLOriginal must contain the source code already
                              SLOriginal    : TSourceList;        // Series of TAnnotatedStringList objects. First (index 0) containes original source code.
                                                                  // The others contain original source code of include files.
                              SLProcessed   : TSourceList;        // Series of TAnnotatedStringList objects. First (index 0) containes processed source code.
                                                                  // The others contain processed source code of include files.
                              SLSourceRTF   : TSourceList;
                              SLRoutines    : TSourceList;
                              ParseAction   : TParseAction;       // "Todo" list
                        const SearchPath    : string;             // Searchpath to be used for include file location
                          VAR Defines,                            // String with conditional defines (e.g. 'debug,master'). Will be updated during parsing
                              Settings      : string;             // String with compiler switches defines (e.g. 'A+,I-'). Will be updated during parsing
                          VAR ErrMsg        : string): integer;   // Error message string.


FUNCTION GetListOfRoutines(const SourceFileName: string;
                                 SLRoutines    : TSourceList;
                                 SearchPath    : string;             // Searchpath to be used for include file location
                             VAR Defines,                            // String with conditional defines (e.g. 'debug,master'). Will be updated during parsing
                                 Settings        : string;             // String with compiler switches defines (e.g. 'A+,I-'). Will be updated during parsing
                             VAR ErrMsg          : string): integer;

FUNCTION GetSourceRTF(const SourceFileName: string;
                            SourceRTF     : TStringList;
                            SearchPath    : string;             // Searchpath to be used for include file location
                        VAR Defines,                            // String with conditional defines (e.g. 'debug,master'). Will be updated during parsing
                            Settings      : string;             // String with compiler switches defines (e.g. 'A+,I-'). Will be updated during parsing
                        VAR ErrMsg        : string): integer;


FUNCTION IsReservedWord(Token: string): boolean;



CONST
  MaxReservedWords = 67;
var
  ReservedWords : ARRAY[0..MaxReservedWords] OF string =  // Partially from Delphi 7 Help
                   ('and', 'array', 'as', 'asm',
                    'begin', 'case', 'class', 'const',
                    'constructor', 'destructor', 'dispinterface', 'div',
                    'do', 'downto', 'else', 'end',
                    'except', 'exports', 'file', 'finalization',
                    'finally', 'for', 'function', 'goto',
                    'if', 'implementation', 'in', 'inherited',
                    'initialization', 'inline', 'interface', 'is',
                    'label', 'library', 'mod', 'nil',
                    'not', 'object', 'of', 'or',
                    'out', 'packed', 'private', 'procedure', 'program',
                    'property', 'public', 'published', 'raise', 'record', 'repeat',
                    'resourcestring', 'set', 'shl', 'shr',
                    'string', 'then', 'threadvar', 'to',
                    'try', 'type', 'unit', 'until',
                    'uses', 'var', 'while', 'with',
                    'xor');



implementation


FUNCTION IsReservedWord(Token: string): boolean;
VAR
  i : integer;
BEGIN
  Result := false;
  Token := UpperCase(Token);
  FOR i := 0 TO MaxReservedWords DO
    IF Token = UpperCase(ReservedWords[i]) then
      begin
        Result := true;
        EXIT;
      end;
END;


{ ---------= TAnnotatedStringList methods =----------------------------}


procedure TAnnotatedStringList.Free;
begin
  // Paranoia: Clear strings as well. Do not rely on Delphi to do that.
  SourceName := '';
  CalledBy   := '';
  inherited free;
end; {TAnnotatedStringList.Free}


procedure TAnnotatedStringList.Clear;
begin
  SourceName := '';
  CalledBy   := '';
  Line       := 0;
  inherited Clear;
end;



procedure TAnnotatedStringList.assign(Source: TPersistent);
{ - Expand "assign" method to the additional public variables like "SourceName", "CalledBy" and "Line". }
begin
  inherited assign(Source);
  IF Source is TAnnotatedStringList then
    with Source as TAnnotatedStringList DO
      begin
        self.SourceName := SourceName;
        self.CalledBy   := CalledBy;
        self.Line       := Line;
      end;
end; {TAnnotatedStringList.assign}



procedure TAnnotatedStringList.LoadFromStream(Stream: TStream);
{ - Replecement of original method that could not handle character #0 within text files }
var
  Size : integer;
  S    : string;


        { Originally a method from TStrings. But that uses #0 for "end of string", which in Borland Pascal is a valid character. This is not C! }
        procedure SetTextStr(const Value: string; Size: integer);
        var
          P, Start, PEnd: PChar;
          S: string;
        begin
          Clear;
          P := Pointer(Value);
          if P <> nil then
            begin
              PEnd := P;
              Inc(PEnd, Size);
              while P < PEnd do begin
                Start := P;
                while not (P^ in [#10, #13]) and (P < PEnd)
                  do Inc(P);
                SetString(S, Start, P - Start);
                Add(S);
                if P^ = #13 then Inc(P);
                if P^ = #10 then Inc(P);
              end;
            end;
        end;



begin
  BeginUpdate;
  try
    Size := Stream.Size - Stream.Position;
    SetString(S, nil, Size);
    Stream.Read(Pointer(S)^, Size);
    SetTextStr(S, Length(S));  // Usually that would be TStrings.SetTextString. Here replaced by a local routine
  finally
    EndUpdate;
  end;
end;



{ ---------= TSourceList methods =-------------------------------------}

procedure TSourceList.Add(ASL: TAnnotatedStringList);
{ - Just to remind coder only to add objects of type TAnnotatedStringList }
begin
  inherited Add(ASL);
end;


procedure TSourceList.Clear;
{ - Frees all TAnnotatedStringList objects before clearing }
var
  i : integer;
begin
  FOR i := Count-1 DOWNTO 0 DO
    IF assigned(TAnnotatedStringList(Items[i])) then
      begin
        TAnnotatedStringList(Items[i]).Free;
        Items[i] := nil;
      end;
  inherited Clear;
end;


procedure TSourceList.Free;
{ - Ensures that all TAnnotatedStringList objects are freed also }
begin
  Clear; // Frees all Items...
  inherited free;
end;




{ ---------= Local routines =---------------------------------}


FUNCTION AddCompSetting(CS: string; DefList: TStringList; VAR Res: integer; VAR ErrorDirective: string): string;
{-#-----------------------------------------------------------------------------
  #Descr  Adds a compiler setting to the Directives in DefList.
          e.g. (*$I-*) will be added as I- and a I+ will be removed
  #param  CS may contain more than one settings: e.g. 'A+,I-,R+'.
          Only the first will be added.
  #param  DefList List of conditional defines
  #param  Res Error
  #param  ErrorDirective Directive that caused the error
  #result Remainig settings. Only the first setting will be added, the others are returend as result
 -#----------------------------------------------------------------------------}
VAR
  i, Ls    : integer;
  s, neg_s : string;
BEGIN
  Res := 0;
  ErrorDirective := '';
  s := UpperCase(copy(CS+',', 1, Pos(',', CS+',')));
  Result := copy(CS, Length(s)+1, Length(CS));
  delete(s, Length(s), 1);  // Remove ","
  // Determine Position of switch ('+','-') in string.
  i := Pos('-', s);
  IF i = 0 then
    i := MaxLongint;
  LS:= Pos('+', s);
  IF LS = 0 then
    LS := MaxLongint;
  IF i < LS THEN
    LS := i;
  IF LS > Length(s) then
    Ls := Length(s);
  IF Ls > 1 then
    begin
      IF s[1] IN [' ', #9] then  // Compiler directive contains space => ignore remaining
        begin
          Result := '';
          EXIT;
        end;
      IF (Pos(' ', s) > 0) or (Pos(#9, s) > 0) then  // Compiler directive contains space => Error
        begin
          Result := '';
          Res := -1;
          ErrorDirective := s;
          EXIT;
        end;
      IF s[Ls] IN ['+','-'] then
        begin
          s := #1+Copy(s, 1, LS);  // Just to create a difference between compiler switches and directives
          Inc(Ls);
          neg_s := s;
          IF s[Ls] = '-'
            then neg_s[Ls] := '+'
            else neg_s[Ls] := '-';
          IF DefList.Find(neg_s, i) then
            DefList.Delete(i);
          DefList.Add(s)
        end
        else begin           // '+' or '-' missing: Maybe Error or "Align" or "MinEnumSize"
          IF ((s[1] = 'A') and (s[Length(s)] IN ['1','2','4','8'])) or
             ((s[1] = 'Z') and (s[Length(s)] IN ['1','2','4'])) or
             (s = 'APPTYPE') or
             (s = 'WARN') or
             (s = 'YD') or             
             (s = 'ALIGN')
            then begin
              // Skip at the moment....
            end
            else begin
              Result := '';
              ErrorDirective := s;
              Res := -1;
            end;
        end;
    end
    else Result := '';
END; {AddCompSetting}



PROCEDURE AddCompSettings(CS: string; DefList: TStringList; VAR Res: integer; VAR ErrorDirective: string);
{ - Adds a list of Compilersettings. c.f. comment to "AddCompSetting" }
VAR
  st : string;
BEGIN
  Res:= 0;
  ErrorDirective := '';
  st := Trim(CS);
  WHILE (Length(st) > 0) and (Res = 0) DO
    st := AddCompSetting(st, DefList, Res, ErrorDirective);
END; {AddCompSettings}



PROCEDURE AddDefines(D: string; DefList: TStringList);
{ - Adds a compiler directive to TStringList "DefList" }
VAR
  i : integer;
BEGIN
  D := UpperCase(Trim(D));
  if not DefList.Find(D, i) then
    DefList.Add(D);
END; {AddDefines}



PROCEDURE AddCSDefines(D: string; DefList: TStringList);
{ - Adds a list of comma separated compiler directives to TStringList "DefList" }
VAR
  s : string;
BEGIN
  WHILE Length(D) > 0 DO
    BEGIN
      s := copy(D+',', 1, Pos(',', D+','));
      D := copy(D, Length(s)+1, Length(D));
      delete(s, Length(s), 1);  // Remove ","
      AddDefines(s, DefList);
    END;
END; {AddCSDefines}



PROCEDURE CreateErrorMsg(VAR ErrMsg: string; const FileName: string; Row: integer; const Info: string);
{ - Construct an error message string from several different information.
    New text will be appended to ErrMsg }
begin
  IF ErrMsg <> '' then
    ErrMsg := ErrMsg+#10;
  ErrMsg := ErrMsg + ' ' + FileName + ' ';
  ErrMsg := ErrMsg + 'Line ' + IntToStr(Row)+': ' + Info;
end; {CreateErrorMsg}



{ ---------= Main function =--------------------------------------------}



FUNCTION RemoveDeadCodeEx(const SourceFileName: string;           // Name of source file to be processed. If empty SLOriginal must contain the source code already
                          const CalledBySource: string;           // "Calling" source file. Usually empty. Only of interest while processing include files
                                CalledFromLine: integer;          // Line numer of call. Usually undefined. Only of interest while processing include files
                                SLOriginal    : TSourceList;      // Series of TAnnotatedStringList objects. First (index 0) containes original source code.
                                                                  // The others contain original source code of include files.
                                SLProcessed   : TSourceList;      // Series of TAnnotatedStringList objects. First (index 0) containes processed source code.
                                                                  // The others contain processed source code of include files.
                                SLSourceRTF   : TSourceList;
                                SLRoutines    : TSourceList;
                                ParseAction   : TParseAction;     // "Todo" list
                          const SearchPath    : string;           // Searchpath to be used for include file location
                                DefList       : TStringList;      // String list containing all conditional defines and compiler settings. Will be updated during parsing
                            VAR ErrMsg        : string): integer; // Error message string.

{ - SourceFileName  Name of source file to be processed. If empty SLOriginal must contain the source code already
    CalledBySource  "Calling" source file. Usually empty. Only of interest while processing include files
    CalledFromLine  Line numer of call. Usually undefined. Only of interest while processing include files
    SLOriginal      Series of TAnnotatedStringList objects. First (index 0) containes original source code.
                    The others contain original source code of include files.
    SLProcessed     Series of TAnnotatedStringList objects. First (index 0) containes processed source code.
                    The others contain processed source code of include files.
    DelStrings      Set to true to replace all characters within strings with space
    DelComments     Set to true to remove all comments from source code
    DelDirectives   Set to true to process all compiler directives and replace outruled source code by space
    SearchPath      Searchpath to be used for include file location
    DefList         String list containing all conditional defines and compiler settings. Will be updated during parsing
    ErrMsg          Error message string. }
TYPE
  TIfDefStack    = ARRAY OF boolean;
VAR
  CommentFound    : char;        // Describes sort of active comment. #0 means "no comment".
  Ch              : Char;
  CompDir         : boolean;     // Found characters belong to a compiler directive statement
  CompDirHandled  : boolean;
  b               : boolean;
  InsideASM       : boolean;
  CompStart       : integer;     // Offset zur aktuellen Leseposition, ab der Bezeichner der Compilerdirective erwartet wird
  CompStartLine   : integer;
  SLIndex, i      : integer;
  ActiveCDir      : string;      // Last found compiler directive
  ActiveCDirPara  : string;      // Parameter for last found compiler directive
  Row, Col        : integer;     // Active reading position
  IfDefStack      : TIfDefStack;
  ActiveLineLength: integer;
  SourceLines     : integer;
  Level           : integer;
  ActiveChar      : char;
  StringBegin     : boolean;
  ErrorDirective  : string;
  ActiveIdentifier: string;      // Actual token being processed

  UpCaseChar      : ARRAY[char] OF char;
  Keyword         : shortstring;
  KeywordID       : integer;
  LastKeywordID   : integer;
  KeyWordID_ASM   : integer;
  KeyWordID_END   : integer;
  KeyWordIDs      : ARRAY[0..MaxReservedWords] OF integer;
  HLString        : string;
  OutRuled        : boolean;
  LastOutRuled    : boolean;
  CommentItalic   : integer;   // Flag indicating that comment text has been set to italic right now
  ItalicActive    : boolean;   // Flag that italic is activated

  KeyWordID_BEGIN          : integer;
  KeyWordID_CASE           : integer;
  KeyWordID_TRY            : integer;
  KeyWordID_RECORD         : integer;
  KeyWordID_CLASS          : integer;
  KeyWordID_OBJECT         : integer;
  KeyWordID_Of             : integer;
  KeyWordID_ROUTINE        : ARRAY[0..3] OF integer;
  KeyWord_ROUTINE          : ARRAY[0..3] OF string;
  KeyWordID_RoutStart      : ARRAY[0..4] OF integer;
  KeyWord_RoutStart        : ARRAY[0..4] OF string;
  KeywordID_IMPLEMENTATION : integer;
  BuildRoutineStr          : integer;
  RoutineStr               : string;
  InsideImplementation     : boolean;
  LastRoutineDefRow        : integer;
  LastRoutineDefCol        : integer;
  ExtraBlankChar           : integer;
  BlockDepth               : integer;
  ProcPrefix               : string;


        FUNCTION IsKeyWord(ID: integer; Name: string): boolean;
        BEGIN
          Result := (KeyWordID = ID) and (KeyWord = Uppercase(Name));
        END;


        // --- Syntax Highlighting -----------------

        PROCEDURE HLStart;
        { - Initialize RTF export }
        BEGIN
          TAnnotatedStringList(SLSourceRTF.Items[SLIndex]).Add(RTFDefLine1);
          TAnnotatedStringList(SLSourceRTF.Items[SLIndex]).Add(RTFDefLine2);
        END;

        PROCEDURE HLAddText(st: string; CharOffset: integer);
        { - Append string st to HLString. }
        BEGIN
          IF not (PA_BuildRTF IN ParseAction) THEN
            EXIT;
          IF CharOffset = 0
            then HLString := HLString + st
            else Insert(st, HLString, Length(HLString)+CharOffset);
        END;


        PROCEDURE HLAddChar(ch: char);
        { - Add character ch to HLString. Take care of special characters }
        BEGIN
          IF not (PA_BuildRTF IN ParseAction) THEN
            EXIT;
          IF ch IN ['{', '}', '\']
            then HLString := HLString + '\' + ch
            else HLString := HLString + ch;
        END;


        PROCEDURE HLSetBackGround(i: integer);
        { - Add control characters to RTF in order to set a background color index }
        BEGIN
          IF not (PA_BuildRTF IN ParseAction) THEN
            EXIT;
          HLString := HLString + '\highlight'+IntToStr(i)+' ';
        END;


        PROCEDURE HLAddNewLine(st: string);
        { - Appends string st to HLString, then adds HLString to SLSourceRTF }
        BEGIN
          IF not (PA_BuildRTF IN ParseAction) THEN
            EXIT;
          HLString := HLString + st;
          TAnnotatedStringList(SLSourceRTF.Items[SLIndex]).Add(HLString);
          HLString := '';
        END;


        PROCEDURE HLSetStandardColor(CharOffset: integer);
        { - Includes the standard color control characters to HLString }
        BEGIN
          IF not (PA_BuildRTF IN ParseAction) THEN
            EXIT;
          IF InsideASM
            THEN HLAddText('\cf4\i0 ', CharOffset)
            ELSE HLAddText('\cf1\i0 ', CharOffset);
          ItalicActive := false;
        END;


        PROCEDURE HLSetCommentItalic;
        { - Adds the italic control characters to HLString }
        BEGIN
          IF not(OutRuled) then
            BEGIN
              HLAddText('\cf2\i ', 0);
              ItalicActive := true;
              CommentItalic := 1;
            END
            ELSE CommentItalic := -1;
        END;


        // --- Routine detection -------------------------

        PROCEDURE AppendRoutineStr(DelChars: integer);
        VAR
          i  : integer;
          st : string;
        BEGIN
          delete(RoutineStr, Length(RoutineStr)+1-DelChars, DelChars);
          RoutineStr := Trim(RoutineStr);
          WHILE Pos('  ', RoutineStr) > 0 DO
            delete(RoutineStr, Pos('  ', RoutineStr), 1);
          RoutineStr := RoutineStr + '|' + IntToStr(LastRoutineDefRow+1) + '|' +
                                           IntToStr(LastRoutineDefCol+1);
          //--- Check if routine has name (otherwise it is just a declaration) ---
          st := RoutineStr;
          delete(st, 1, Pos('|',st));
          delete(st, 1, Pos('|',st));
          st := UpperCase(st);
          FOR i := 0 TO 3 DO
            IF Pos(KeyWord_ROUTINE[i], st) = 1 then
              begin
                delete(st, 1, Length(KeyWord_ROUTINE[i]));
                st := Trim(st);
                break;
              end;
          // --- Add to list of routines, if not a declaration ---
          IF not(st[1] IN [':', '=', '(', ';']) THEN
            BEGIN
              TAnnotatedStringList(SLRoutines.Items[SLIndex]).Add(RoutineStr);
              delete(st, Pos('|', st), Length(st));
              ProcPrefix := ProcPrefix + '.' + st;
            END;

          RoutineStr :='';
          BuildRoutineStr := 0;
        END;

        PROCEDURE CheckRoutines;
        VAR
          i : integer;
        BEGIN
          IF not(InsideIMPLEMENTATION) then
            exit;
          // --- Start of a new routine? ----
          FOR i := 0 TO 3 DO
            IF IsKeyWord(KeyWordID_ROUTINE[i], KeyWord_ROUTINE[i]) then
              BEGIN
                IF (RoutineStr <> '')
                  then AppendRoutineStr(Length(KeyWord_ROUTINE[i]));
                RoutineStr := IntToStr(Row+1)+'|'+IntToStr(Col - Length(KeyWord_ROUTINE[i]))+'|'+KeyWord_ROUTINE[i] + ' ';
                BuildRoutineStr := 1;
                EXIT;
              END;
          // --- Routine has body? (=real?) ----
          IF (RoutineStr <> '') and (BuildRoutineStr < 2) then
            FOR i := 0 TO 4 DO
              IF IsKeyWord(KeyWordID_ROUTStart[i], KeyWord_ROUTStart[i]) then
                AppendRoutineStr(Length(KeyWord_ROUTStart[i]));
        END;


        PROCEDURE AddCharToRoutineStr(ch: char); 
        { Adds a character to RoutineStr }
        BEGIN
          IF BuildRoutineStr >= 1 THEN
            BEGIN
              RoutineStr := RoutineStr + ch;
              IF (ch = '(') then
                Inc(BuildRoutineStr);
              IF (ch = ')') then
                Dec(BuildRoutineStr);
              IF (ch = ';') THEN
                BEGIN
                  LastRoutineDefRow := Row;
                  LastRoutineDefCol := Col;
                END;
              {
              IF (ch = ';') and (BuildRoutineStr = 1) then
                BuildRoutineStr := 0;
              }
            END;
        END;

        // --- Keyword detection -------------------------
        PROCEDURE CheckBlock;
        BEGIN
          IF IsKeyWord(KeyWordID_BEGIN, 'BEGIN') or
             IsKeyWord(KeyWordID_ASM, 'ASM') then
            begin
              Inc(BlockDepth);
              IF Pos('.', ProcPrefix) > 0 then
                delete(ProcPrefix, Pos('.', ProcPrefix), Length(ProcPrefix));
              exit;
            end;
          IF IsKeyWord(KeyWordID_TRY, 'TRY') or
             IsKeyWord(KeyWordID_RECORD, 'RECORD') or
             IsKeyWord(KeyWordID_CLASS, 'CLASS') then
            begin
              Inc(BlockDepth);
              exit;
            end;
          IF IsKeyWord(KeyWordID_OBJECT, 'OBJECT') then
            IF LastKeywordID <> KeyWordID_Of then  // Todo: Not correct! Object-Name could have same ID as term "of"
              begin
                Inc(BlockDepth);
                exit;
              end;
          IF (BlockDepth > 0) and IsKeyWord(KeyWordID_CASE, 'CASE') then
            Inc(BlockDepth);
          IF (BlockDepth > 0) and IsKeyWord(KeyWordID_END, 'END') then
            Dec(BlockDepth);
        END;


        PROCEDURE ResetKeyWord;
        { Reset Keyword and Keyword-ID. }
        BEGIN
          Keyword       := '';
          LastKeywordID := KeywordID;
          KeyWordID     := 0;
        END;

        PROCEDURE AddCharToKeyword(ch: char);
        { Adds a character to keyword. Updates KeyWordID }
        BEGIN
          ch := UpCaseChar[ch];
          KeyWord := KeyWord + ch;
          KeyWordID := (KeyWordID shl 4) xor byte(ch);
        END;

        PROCEDURE CheckKeyWordForASMBlock;
        { Check if inside/outside ASM block }
        BEGIN
          IF Keyword = '' THEN
            EXIT;
          CheckBlock;
          IF InsideASM and IsKeyWord(KeyWordID_END, 'END') THEN
            BEGIN
              InsideASM := false;
              HLSetStandardColor(-3);
            END;
          IF IsKeyWord(KeyWordID_ASM, 'ASM') THEN
            BEGIN
              InsideASM := true;
              HLSetStandardColor(0);
              IF (PA_ListRoutines IN ParseAction) THEN
                CheckRoutines;
            END;
        END;


        PROCEDURE HandleKeyWord;
        VAR
          i :integer;
        BEGIN
          CheckKeyWordForASMBlock;

          IF (PA_BuildRTF IN ParseAction) and
             not (InsideASM or OutRuled) THEN
            BEGIN
              FOR i := 0 TO MaxReservedWords DO
                IF IsKeyWord(KeyWordIDs[i], ReservedWords[i]) THEN
                  BEGIN
                    HLAddText('\b ', 1-Length(KeyWord));
                    HLAddText('\b0 ', 0);
                    break;
                  END;
            END;

          IF (PA_ListRoutines IN ParseAction) and
             not (InsideASM or OutRuled) THEN
            begin
              IF IsKeyword(KeywordID_IMPLEMENTATION, 'IMPLEMENTATION') then
                InsideImplementation := true;
              CheckRoutines;
            end;

          ResetKeyword;
        END;


        FUNCTION GetKeyWordID(Token: string): integer;
        { Calculate KeywordID for a given string }
        VAR
          i :integer;
        BEGIN
          Result := 0;
          FOR i := 1 TO Length(Token) DO
            Result := (Result shl 4) xor byte(UpCaseChar[Token[i]]);
        END;


        // --- Help routines ---------------------------

        FUNCTION GetChar(AddPos: integer): char;
        { - Returns character at position SL[Row][Col+AddPos]. Returns #0 if position is not valid }
        BEGIN
          IF (Length(TAnnotatedStringList(SLOriginal[SLIndex]).Strings[Row]) >= Col+AddPos) and
             (Col+AddPos > 0)
            then Result := TAnnotatedStringList(SLOriginal[SLIndex]).Strings[Row][Col+AddPos]
            else Result := #0;
        END;



        FUNCTION CompDirValid: boolean;
        { - Returns false, if active editor position is ruled out by compiler directive }
        BEGIN
          Result := true;
          IF High(IfDefStack) >= 0 then
            Result := IfDefStack[High(IfDefStack)];
        END;


        PROCEDURE PushIfDefStack(b: boolean);
        { - Adds a compiler directive to the stack (ifdef, endif) }
        BEGIN
          SetLength(IfDefStack, Length(IfDefStack)+1);  // Sorry for that, poor memory manager...
          IfDefStack[High(IfDefStack)] := b;
        END;



        PROCEDURE BlankActiveChar;
        { - Sets character at position Col,Row in StringList SLProcessed to #32 }
        VAR
          st : string;
        BEGIN
          st := TAnnotatedStringList(SLProcessed.Items[SLIndex]).Strings[Row];
          st[Col] := ' ';
          TAnnotatedStringList(SLProcessed.Items[SLIndex]).Strings[Row] := st;
          ActiveChar := ' ';
        END;

        PROCEDURE BlankPreviousActiveChar;
        { - Sets character at position Col-1,Row in StringList SLProcessed to #32 }
        VAR
          st : string;
        BEGIN
          IF Col > 0 then
            begin
              st := TAnnotatedStringList(SLProcessed.Items[SLIndex]).Strings[Row];
              st[Col-1] := ' ';
              TAnnotatedStringList(SLProcessed.Items[SLIndex]).Strings[Row] := st;
              ActiveChar := ' ';
            end;
        END;



        FUNCTION KnownDirective(DirectiveStr: string): boolean;
        { - Checks wether compiler directive stored in ActiveCDir is identical
            to DirectiveStr. If so, CompDirHandled is set to true }
        BEGIN
          Result := (DirectiveStr = ActiveCDir);
          IF Result then
            CompDirHandled := true;
        END;


        FUNCTION InvertCompDir(CD: string): string;
        { - Toggles Compiler directive:  $A+ will change to $A- }
        BEGIN
          Result := '';
          IF length(CD) < 1 then
            EXIT;
          IF CD[Length(CD)] = '-'
            THEN CD[Length(CD)] := '+'
            ELSE IF CD[Length(CD)] = '+'
              THEN CD[Length(CD)] := '-';
          Result := CD;
        END;



        FUNCTION ProcessInclude(FileName: string): integer;
        { Handles recursive procedure call to process an include file }
        BEGIN
          Result := -1;
          if ExtractFileExt(FileName) = '' then
            ChangeFileExt(FileName, '.pas');
          FileName := FileSearch(FileName, SearchPath);
          IF FileName = '' then
            EXIT;
          Result := RemoveDeadCodeEx(FileName, SourceFileName, Row, SLOriginal, SLProcessed, nil, SLRoutines, ParseAction, SearchPath, DefList, ErrMsg);
        END; {ProcessInclude}


        PROCEDURE SkipMinorCompilerDirectives;
        { Skip compiler directives that (hopefully) do not affect code analysis }
        CONST
          MaxSkipList = 25;
          SkipList : ARRAY[0..MaxSkipList] OF string =
                 ('WARNINGS',
                  'DEBUGINFO',
                  'DESCRIPTION',
                  'EXTENSION',
                  'PRELOAD',
                  'IMAGEBASE',
                  'IMPLICITBUILD',
                  'NOINCLUDE',
                  'NODEFINE',
                  'EXTERNALSYM',
                  'HPPEMIT',
                  'DESIGNONLY',
                  'DENYPACKAGEUNIT',
                  'RUNONLY',
                  'WEAKPACKAGEUNIT',
                  'ASSERTIONS',
                  'APPTYPE',
                  'BOOLEVAL',
                  'LIBPREFIX',
                  'SOPREFIX',
                  'LIBSUFFIX',
                  'LIBVERSION',
                  'MINENUMSIZE',
                  'LONGSTRINGS',
                  'REALCOMPATIBILITY',
                  'OBJEXPORTALL');
        VAR
          i : integer;
        BEGIN
          IF CompDirHandled THEN
            EXIT;
          FOR i := 0 TO MaxSkipList DO
            IF KnownDirective(SkipList[i]) THEN
              break; // CompDirHandled is already set to true by KnownCompilerDirective
        END;


        FUNCTION HandleFatCompilerDirectives(Params: string): integer;
        { Handle compiler directives like (*$IOCHECKS ON*). Does so by replacing them, e.g. to (*$I+*) }
        TYPE
          TDirective= RECORD
                        Long  : string;
                        Short : char;
                      END;
        CONST
          MaxDirective = 20;
          Directive : ARRAY[0..MaxDirective] OF TDirective =
                       ((Long:'ALIGN';          Short:'A'),  // Todo: Handle A1, A4 etc...
                        (Long:'BOOLEVAL';       Short:'B'),
                        (Long:'ASSERTIONS';     Short:'C'),
                        (Long:'DEBUGINFO';      Short:'D'),
                        (Long:'IMPORTEDDATA';   Short:'G'),
                        (Long:'LONGSTRINGS';    Short:'H'),
                        (Long:'IOCHECKS';       Short:'I'),
                        (Long:'WRITEABLECONST'; Short:'J'),
                        (Long:'LOCALSYMBOLS';   Short:'L'),
                        (Long:'TYPEINFO';       Short:'M'),
                        (Long:'OPTIMIZATION';   Short:'O'),
                        (Long:'OPENSTRINGS';    Short:'P'),
                        (Long:'OVERFLOWCHECKS'; Short:'Q'),
                        (Long:'RANGECHECKS';    Short:'R'),
                        (Long:'TYPEDADDRESS';   Short:'T'),
                        (Long:'SAFEDIVIDE';     Short:'U'),
                        (Long:'STACKFRAMES';    Short:'W'),
                        (Long:'EXTENDEDSYNTAX'; Short:'X'),
                        (Long:'REFERENCEINFO';  Short:'Y'),
                        (Long:'HINTS';          Short:' '), {Suggested by Gruber_Hans_12345}
                        (Long:'STACKCHECKS';    Short:' '));
        VAR
          ShortForm : char;
          ShortVal  : char;
          i         : integer;
        BEGIN
          Result := 0;
          ShortForm := #0;
          FOR i := 0 TO MaxDirective DO
            IF KnownDirective(Directive[i].Long) THEN
              BEGIN
                ShortForm := Directive[i].Short;
                break;
              END;
          IF ShortForm <> #0 then
            begin
              ShortVal := ' ';
              Params := Uppercase(Params);
              IF Params = 'ON'
                then ShortVal := '+';
              IF Params = 'OFF'
                then ShortVal := '-';
              if (ShortVal = ' ') and (ShortForm = 'A') then
                begin                     // Treat $A2, $A4, $A8 like $A+. Is that the right thing to do???
                  IF (Params = '2') or (Params = '4') or (Params = '8')
                    then ShortVal := '+';
                  IF (Params = '1')
                    then ShortVal := '-';
                end;
              AddCompSettings(ShortForm+ShortVal, DefList, Result, ErrorDirective);
              IF Result <> 0 then
                CreateErrorMsg(ErrMsg, SourceFileName, Row, 'Unknown compiler directive ' + ErrorDirective);
            end;
        END;




begin
  Level          := 0;
  ErrMsg         := '';
  Result         := 0;
  InsideASM      := false;
  OutRuled       := false;
  ItalicActive   := false;
  ExtraBlankChar := 0;
  FOR ch := #0 TO #255 DO
    UpCaseChar[ch] := UpCase(ch);

  KeyWord_ROUTINE[0] := 'PROCEDURE';
  KeyWord_ROUTINE[1] := 'FUNCTION';
  KeyWord_ROUTINE[2] := 'DESTRUCTOR';
  KeyWord_ROUTINE[3] := 'CONSTRUCTOR';
  FOR i := 0 TO 3 DO
    KeyWordID_ROUTINE[i] := GetKeyWordID(KeyWord_ROUTINE[i]);

  KeyWord_RoutStart[0] := 'BEGIN';
  KeyWord_RoutStart[1] := 'VAR';
  KeyWord_RoutStart[2] := 'TYPE';
  KeyWord_RoutStart[3] := 'CONST';
  KeyWord_RoutStart[4] := 'ASM';
  FOR i := 0 TO 4 DO
    KeyWordID_ROUTStart[i] := GetKeyWordID(KeyWord_ROUTStart[i]);

  KeyWordID_ASM         := GetKeyWordID('ASM');
  KeyWordID_END         := GetKeyWordID('END');
  KeyWordID_TRY         := GetKeyWordID('TRY');
  KeyWordID_CASE        := GetKeyWordID('CASE');
  KeyWordID_RECORD      := GetKeyWordID('RECORD');
  KeyWordID_CLASS       := GetKeyWordID('CLASS');
  KeyWordID_OBJECT      := GetKeyWordID('OBJECT');
  KeyWordID_OF          := GetKeyWordID('OF');  
  KeyWordID_BEGIN       := GetKeyWordID('BEGIN');
  KeywordID_IMPLEMENTATION := GetKeyWordID('IMPLEMENTATION');
  ResetKeyWord;
  BuildRoutineStr       := 0;
  InsideImplementation  := false;
  LastKeywordID := 0;
  ProcPrefix := '';

  IF PA_BuildRTF IN ParseAction THEN
    FOR i := 0 TO MaxReservedWords DO
      BEGIN
        ReservedWords[i] := UpperCase(ReservedWords[i]);
        KeyWordIDs[i] := GetKeyWordID(UpperCase(ReservedWords[i]));
      END;

  IF SourceFileName = '' then
    begin  // No file name available? In this case SLOriginal is assumed to contain the source code to be processed already
      IF SLOriginal.Count <> SLProcessed.Count+1
        THEN BEGIN
          ErrMsg := 'Call without source file name specified. Source code should be delivered as parameter!';
          Result := -1;
          EXIT;
        END;
      SLIndex := SLOriginal.Count-1;
    end
    else begin
      SLIndex := SLOriginal.Count;
      SLOriginal.Add(TAnnotatedStringList.Create);
      TAnnotatedStringList(SLOriginal.Items[SLIndex]).Clear;
      TAnnotatedStringList(SLOriginal.Items[SLIndex]).LoadFromFile(SourceFileName);
    end;

  TAnnotatedStringList(SLOriginal.Items[SLIndex]).SourceName := SourceFileName;
  TAnnotatedStringList(SLOriginal.Items[SLIndex]).CalledBy   := CalledBySource;
  TAnnotatedStringList(SLOriginal.Items[SLIndex]).Line       := CalledFromLine;
  SLProcessed.Add(TAnnotatedStringList.Create);
  TAnnotatedStringList(SLProcessed.Items[SLIndex]).Assign(TAnnotatedStringList(SLOriginal.Items[SLIndex]));
  SourceLines := TAnnotatedStringList(SLOriginal.Items[SLIndex]).Count;

  if TAnnotatedStringList(SLOriginal.Items[SLIndex]).count < 1 then  { empty source? }
    exit;

  IF PA_BuildRTF IN ParseAction THEN
    IF @SLSourceRTF = nil then
      ParseAction := ParseAction - [PA_BuildRTF];

  IF PA_ListRoutines IN ParseAction THEN
    IF @SLRoutines = nil then
      ParseAction := ParseAction - [PA_ListRoutines];

  IF PA_BuildRTF IN ParseAction THEN
    BEGIN
      SLSourceRTF.Add(TAnnotatedStringList.Create);
      HLStart;
    END;

  IF PA_ListRoutines IN ParseAction THEN
    BEGIN
      SLRoutines.Add(TAnnotatedStringList.Create);
    END;

  SetLength(IfDefStack, 0);

  Row := 0;
  Col := 0;
  CommentFound     := #0;
  CompDir          := false;
  CompDirHandled   := false;
  CompStart        := 0;
  CompStartLine    := 0;
  ActiveLineLength := Length(TAnnotatedStringList(SLOriginal.Items[SLIndex]).Strings[Row]);
  StringBegin      := false;
  ActiveIdentifier := '';
  BlockDepth       := 0;

  REPEAT
    // --- Search next valid character (any) ---
    Repeat
      Inc(Col);
      IF Col > ActiveLineLength then
        begin
          Col := 0;
          Inc(Row);
          { -- Insert following to check Blockdepth algorithm:
          TAnnotatedStringList(SLProcessed.Items[0]).strings[Row-1] :=
                    TAnnotatedStringList(SLProcessed.Items[0]).strings[Row-1] + ' ' + IntToStr(BlockDepth);
          IF BuildRTF THEN
            TAnnotatedStringList(SLSourceRTF.Items[0]).strings[Row-1] :=
              TAnnotatedStringList(SLSourceRTF.Items[0]).strings[Row-1] + ' ' + ProcPrefix+' ' + IntToStr(BlockDepth);
          {}
          if (Row < SourceLines) then
            ActiveLineLength := Length(TAnnotatedStringList(SLOriginal.Items[SLIndex]).Strings[Row]);
          IF CommentFound = '/' then
            begin
              IF not(OutRuled) then
                BEGIN
                  HLAddText('\cf1\i0 ', 0);
                  ItalicActive := false;
                END;
              CommentFound := #0;
            end;
          IF (CommentFound = '''') or (CommentFound = '"') then  // Unterminated string
            begin
              CreateErrorMsg(ErrMsg, SourceFileName, Row, 'Unterminated string');
              Result := -2;
            end;
          HandleKeyWord;
          IF ItalicActive or true
            then HLAddNewLine('\par')
            else HLAddNewLine('\i  \i0\par');  //Just add italc space to give all lines the same height
        end;
     until (Row >= SourceLines) or (Col > 0);

     if (Row >= SourceLines) then
       break;


     // --- Check for start of new comment or string ---
     CommentItalic := 0;
     Ch := GetChar(0);
     IF (Row < SourceLines) and (CommentFound = #0) then
       begin
         StringBegin := false;
         CASE Ch OF
           '''' : begin
                    CommentFound := '''';
                    StringBegin  := true;
                    IF not(OutRuled) then
                      HLSetBackGround(5);
                  end;
           '{'  : begin
                    CommentFound  := '{';
                    CompDir := GetChar(1) = '$';
                    CompStart := 2;
                    HLSetCommentItalic;
                    IF (PA_InvDelComments IN ParseAction) then
                      ExtraBlankChar := 1;
                  end;
           '('  : IF GetChar(1) = '*' then
                    begin
                      CommentFound := '(';
                      CompDir := GetChar(2) = '$';
                      CompStart := 3;
                      HLSetCommentItalic;
                      IF (PA_InvDelComments IN ParseAction) then
                        ExtraBlankChar := 2;
                    end;
           '/'  : IF GetChar(1) = '/'
                    then begin
                      CommentFound := '/';
                      HLSetCommentItalic;
                    end;
           '"'  : IF InsideASM THEN
                    begin
                      CommentFound := '"';
                      StringBegin  := true;
                      IF not OutRuled then
                        HLSetBackGround(5);
                    end;
         end; {case}
         IF CommentFound <> #0 then
           begin
             CompStartLine := Row+1;
           end;

         ActiveChar := ch;
         IF ch IN ['0'..'9', 'A'..'Z', 'a'..'z', '_']
           THEN AddCharToKeyWord(ch)
           ELSE HandleKeyWord;

         // --- If comment is a Compilerdirective, get information about it ---
         IF CompDir and (PA_DelDirectives IN ParseAction) then
           begin
             ActiveCDir := '';
             ActiveCDirPara := '';
             i := CompStart;
             WHILE not(GetChar(CompStart) IN [#0, '}', '*', '{', '(']) DO
               Inc(CompStart);
             ActiveCDir := copy(TAnnotatedStringList(SLOriginal[SLIndex]).Strings[Row], Col+i, CompStart-i);
             ActiveCDir := AnsiReplaceText(ActiveCDir, #9, ' ');
             ActiveCDir := Uppercase(ActiveCDir);
             CompDirHandled := false;
             IF (Length(ActiveCDir) > 1) and (ActiveCDir[1] <> ' ') then  // No space between $ and compiler directive name!
               begin
                 IF Pos(' ', ActiveCDir) > 1 then  // Compiler-Directive incl. parameter?
                   begin
                     ActiveCDirPara := Trim(copy(ActiveCDir, Pos(' ', ActiveCDir)+1, 1024));
                     delete(ActiveCDir, Pos(' ', ActiveCDir), Length(ActiveCDir));
                     IF CompDirValid then
                       begin
                         IF KnownDirective('IFOPT')
                           then begin
                             b := DefList.Find(#1+ActiveCDirPara, i);
                             PushIfDefStack(b);
                             IF not b THEN
                               BEGIN
                                 b := DefList.Find(InvertCompDir(#1+ActiveCDirPara), i);  // Check, if inverted compiler directive is known
                                 IF (not b) and StrictDefCheck then
                                   begin
                                     Result := -4;
                                     CreateErrorMsg(ErrMsg, SourceFileName, Row, 'Compiler switch '+ copy(ActiveCDirPara, 1, Length(ActiveCDirPara)-1) + ' not specified');
                                   end;
                               END;
                           end
                           else IF KnownDirective('IFDEF')
                             then PushIfDefStack(DefList.Find(ActiveCDirPara, i))
                             else IF KnownDirective('IFNDEF')
                               then PushIfDefStack(not DefList.Find(ActiveCDirPara, i))
                               else IF KnownDirective('DEFINE') then
                                 begin
                                   if not DefList.Find(ActiveCDirPara, i)
                                     then DefList.Add(ActiveCDirPara);
                                 end
                                 else IF KnownDirective('UNDEF') then
                                   if DefList.Find(ActiveCDirPara, i) then
                                     DefList.Delete(i);
                       end
                       else begin  // Source ruled out by compiler directives?
                         IF KnownDirective('IFOPT') or
                            KnownDirective('IFDEF') or
                            KnownDirective('IFNDEF') then
                           Inc(Level);
                       end;
                   END;
                 IF KnownDirective('ELSE') and (Level = 0) THEN
                   IF Length(IfDefStack) > 0 then
                     IfDefStack[High(IfDefStack)] := not(IfDefStack[High(IfDefStack)]);
                 IF KnownDirective('ENDIF') THEN
                   IF level = 0
                     then begin
                       IF Length(IfDefStack) > 0 then
                         SetLength(IfDefStack, Length(IfDefStack)-1);
                     end
                     else Dec(Level);

                 IF KnownDirective('I') or KnownDirective('INCLUDE') THEN { Include files }
                   Result := ProcessInclude(ActiveCDirPara); // Recursive call!

                 IF not CompDirHandled and CompDirValid then  // Maybe "{$WARNINGS}" or something else
                   SkipMinorCompilerDirectives;

                 IF not CompDirHandled and CompDirValid then  // Maybe "fat" compiler directive like "IOCHECKS ON"
                   Result := HandleFatCompilerDirectives(ActiveCDirPara);

                 IF not CompDirHandled and CompDirValid then  // Maybe simple define of sort "{$I+}"
                   BEGIN
                     ActiveCDirPara := copy(ActiveCDir, Pos('$', ActiveCDir)+1, Length(ActiveCDir));
                     AddCompSettings(ActiveCDirPara, DefList, Result, ErrorDirective);
                     IF Result <> 0 then
                       CreateErrorMsg(ErrMsg, SourceFileName, Row, 'Unknown compiler directive ' + ErrorDirective);
                   END;
               end;
           end
           else begin
             CompDirHandled := false;
           end;
       END;

     // --- Remove strings from SL2 ---
     IF (CommentFound = '''') and (not CompDir) and (PA_DelStrings IN ParseAction) then
       IF GetChar(0) <> '''' then
         BlankActiveChar;
     IF (CommentFound = '"') and (not CompDir) and (PA_DelStrings IN ParseAction) then
       IF GetChar(0) <> '"' then
         BlankActiveChar;

     // --- Remove comment from SL2 ---
     IF not(CommentFound IN [#0, '''', '"']) and (not CompDir) and (PA_DelComments IN ParseAction) then
       BlankActiveChar;

     // --- Remove "inverse" comment from SL2 ---
     IF ((CommentFound IN [#0, '''', '"']) or (CompDir)) and (PA_InvDelComments IN ParseAction) then
       BlankActiveChar;

     IF ExtraBlankChar > 0 then
       BEGIN
         Dec(ExtraBlankChar);
         BlankActiveChar;
       END;

     // --- Remove deactivated code (by compiler directive) from SL2 ---
     LastOutRuled := OutRuled;
     OutRuled := (not(CompDirValid) or CompDirHandled);
     IF OutRuled and (PA_DelDirectives IN ParseAction) THEN
       BlankActiveChar;
     IF (LastOutRuled <> OutRuled) then
       begin
         IF OutRuled
           then HLSetBackground(6)
           else begin
             IF CommentItalic = 1
               then HLAddText('\cf1 ', 0)
               else begin
                 HLAddText('\cf1\i0 ', 0);
                 ItalicActive := false;
               end;
             HLSetBackground(0);
             IF CommentItalic = -1
               THEN HLSetCommentItalic;
           end;
       end;
     HLAddChar(ch);
     AddCharToRoutineStr(ActiveChar);

     // --- Check for end of comment ---
     IF (CommentFound <> #0) then
       begin
         CASE CommentFound OF
           '''' : IF (GetChar(0) = '''') and not StringBegin then
                    begin
                      CommentFound := #0;
                      IF not(OutRuled) then
                        HLSetBackGround(0);
                    end;
           '{'  : IF GetChar(0) = '}' then
                    begin
                      CommentFound := #0;
                      IF not(OutRuled) then
                        BEGIN
                          HLAddText('\cf1\i0 ', 0);
                          ItalicActive := false;
                        ENd;
                      IF (PA_InvDelComments IN ParseAction) then
                        BlankActiveChar;
                    end;
           '('  : IF (GetChar(0) = ')') and (GetChar(-1) = '*') then
                    begin
                      CommentFound := #0;
                      IF not(OutRuled) then
                        BEGIN
                          HLAddText('\cf1\i0 ', 0);
                          ItalicActive := false;
                        ENd;
                      IF (PA_InvDelComments IN ParseAction) then
                        BEGIN
                          BlankActiveChar;
                          BlankPreviousActiveChar;
                        END;

                    end;
           '"' : IF (GetChar(0) = '"') and not StringBegin then
                   begin
                     CommentFound := #0;
                     IF not OutRuled then
                       HLSetBackGround(0);
                   end;
         END; {case}
         CompDir := CompDir and (CommentFound <> #0);
         IF not(CompDir) then
           CompDirHandled := false;
       end;
    StringBegin := false;

  UNTIL false or (Result <> 0);
  HLAddNewLine('}');  // Close RTF

  IF CommentFound <> #0 then
    begin
      Result := -3;
      CreateErrorMsg(ErrMsg, SourceFileName, Row, 'Unexpected end of file in comment started on line ' + IntToStr(CompStartLine));
    end;

end; {RemoveDeadCodeEx}




{ ---------= Simplified call of the main function =---------------------}



FUNCTION RemoveDeadCode(const SourceFileName: string;             // Name of source file to be processed. If empty, SLOriginal must contain the source code already
                              SLOriginal    : TSourceList;        // Series of TAnnotatedStringList objects. First (index 0) containes original source code.
                                                                  // The others contain original source code of include files.
                              SLProcessed   : TSourceList;        // Series of TAnnotatedStringList objects. First (index 0) containes processed source code.
                                                                  // The others contain processed source code of include files.
                              SLSourceRTF   : TSourceList;
                              SLRoutines    : TSourceList;
                              ParseAction   : TParseAction;       // "Todo" list while parsing
                        const SearchPath    : string;             // Searchpath to be used for include file location
                          VAR Defines,                            // String with conditional defines (e.g. 'debug,master'). Will be updated during parsing
                              Settings      : string;             // String with compiler switches defines (e.g. 'A+,I-'). Will be updated during parsing
                          VAR ErrMsg        : string): integer;   // Error message string.
{ - Removes Comments and Compiler-Directives (incl. their dead code) from pascal source in source code defined by SourceFileName.
    SourceFileName contains the file name of the source file to be analyzed.
    If SourceFileName is empty, SLOriginal is assumed to contain the source to be analyzed
    within the last TAnnotatedStringList object. In that case SLOriginal should contain one element more than SLProcessed.
    Defines contains comma separated defines, e.g. 'MSWindows,debug'. String will be updated to reflect changes found while processing the file
    Settings contains comma separated compiler switches, e.g. 'A+,I-,L-' String will be updated to reflect changes found while processing the file
    ErrMsg returns description for error found
    SLOriginal.Items[0] returns original code
    SLOriginal.Items[1] and following contain original file code...
    SLProcessed.Items[0] returns extracted, processed code
    SLProcessed.Items[1] and following contain processed include file code... }
VAR
  i               : integer;
  DefList         : TStringList;
  ErrorDefine, st : string;


        PROCEDURE AddCSDefines(D: string);
        { - Adds a list of comma separated compiler directives to TStringList DefList }
        VAR
          s : string;
        BEGIN
          WHILE Length(D) > 0 DO
            BEGIN
              s := copy(D+',', 1, Pos(',', D+','));
              D := copy(D, Length(s)+1, Length(D));
              delete(s, Length(s), 1);  // Remove ","
              AddDefines(s, DefList);
            END;
        END; {AddCSDefines}


BEGIN
  ErrMsg := '';
  Result := 0;
  DefList := TStringList.Create;
  DefList.Sorted := true;

  AddCSDefines(Defines);
  AddCompSettings(Settings, DefList, Result, ErrorDefine);
  IF Result <> 0 then
    CreateErrorMsg(ErrMsg, SourceFileName, -1, 'Unknown compiler directive ' + ErrorDefine);

  IF Result = 0 then
    Result := RemoveDeadCodeEx(SourceFileName, '', -1, SLOriginal, SLProcessed, SLSourceRTF, SLRoutines, ParseAction, SearchPath, DefList, ErrMsg);

  // --- Return actual Defines and Settings as string ----
  Settings := '';
  Defines := '';
  FOR i := 0 TO DefList.Count-1 DO
    BEGIN
      st := DefList.Strings[i];
      IF Length(st) > 0 then
        begin
          IF st[1] = #1 then
            begin
              delete(st, 1, 1);
              Settings := Settings + ',' + st;
            end
            else begin
              Defines := Defines + ',' + st;
            end;
        end;
    END;
  delete(Settings, 1, 1);
  delete(Defines, 1, 1);
  DefList.Free;
  DefList := nil;
END; {RemoveDeadCode}




FUNCTION GetListOfRoutines(const SourceFileName: string;
                                 SLRoutines    : TSourceList;
                                 SearchPath    : string;             // Searchpath to be used for include file location
                             VAR Defines,                            // String with conditional defines (e.g. 'debug,master'). Will be updated during parsing
                                 Settings      : string;             // String with compiler switches defines (e.g. 'A+,I-'). Will be updated during parsing
                             VAR ErrMsg        : string): integer;
VAR
  SLOriginal,
  SLProcessed  : TSourceList;
begin
  SLOriginal  := TSourceList.Create;
  SLProcessed := TSourceList.Create;

  Result := RemoveDeadCode(SourceFileName,
                           SLOriginal,
                           SLProcessed,
                           nil,
                           SLRoutines,
                           [PA_DelStrings, PA_DelComments, PA_DelDirectives, PA_ListRoutines],
                           SearchPath,
                           Defines,                            // String with conditional defines (e.g. 'debug,master'). Will be updated during parsing
                           Settings,
                           ErrMsg);
  SLProcessed.Free;
  SLOriginal.Free;
END;




FUNCTION GetSourceRTF(const SourceFileName: string;
                            SourceRTF     : TStringList;
                            SearchPath    : string;             // Searchpath to be used for include file location
                        VAR Defines,                            // String with conditional defines (e.g. 'debug,master'). Will be updated during parsing
                            Settings      : string;             // String with compiler switches defines (e.g. 'A+,I-'). Will be updated during parsing
                        VAR ErrMsg        : string): integer;
VAR
  SLOriginal,
  SLProcessed  : TSourceList;
  SLRTF        : TSourceList;
begin
  SourceRTF.Clear;
  SLOriginal  := TSourceList.Create;
  SLProcessed := TSourceList.Create;
  SLRTF       := TSourceList.Create;

  Result := RemoveDeadCode(SourceFileName,
                           SLOriginal,
                           SLProcessed,
                           SLRTF,
                           nil,
                           [PA_BuildRTF],
                           SearchPath,
                           Defines,                            // String with conditional defines (e.g. 'debug,master'). Will be updated during parsing
                           Settings,
                           ErrMsg);

  IF SLProcessed.Count > 0 then
    SourceRTF.Assign(TStringList(SLRTF[0]));
  SLRTF.Free;
  SLProcessed.Free;
  SLOriginal.Free;
END;





end.

