unit DotGlob;

{*****************************************************************}
{                                                                 }
{  DotGlob                                                        }
{                                                                 }
{  Communication with DOT.exe from GraphViz                       }
{  For GraphViz visit www.graphviz.org                            }
{                                                                 }
{  Translates unit names and relations between that units into    }
{  the DOT language. Then calls DOT with that file in order to    }
{  get a beautiful graph                                          }
{                                                                 }
{  (c) 2007, 2008 M. Braun                                        }
{  Contact: info@grizzlymotion.com                                }
{                                                                 }
{*****************************************************************}



interface


USES Windows, Classes, Graphics, SysUtils, Types, StrUtils, Math, Exec, JPEG, ComCtrls, Forms, ClipBrd;


TYPE
  TBoxType  = (BT_None, BT_Box, BT_Diamond, BT_Circle, BT_Ellipse, BT_Hexagon, BT_Triangle, BT_InvTriangle, BT_Trapezium, BT_Parallelogram);

CONST
  CBoxTypeName: ARRAY[TBoxType] OF string = ('none', 'box', 'diamond', 'circle', 'ellipse', 'hexagon', 'triangle', 'invtriangle', 'trapezium', 'parallelogram');

CONST
  CMaxBPoint  = 255;
  CPixPerInch = 96;

TYPE
  TPointSgl   = RECORD
                    X, Y : single;
                END;
  TBPointArr  = ARRAY[0..CMaxBPoint] OF TPointSgl;  // Room for 21 groups of points
  TPTBPointArr= ^TBPointArr;
  TDotNode    = RECORD
                  Identifier        : integer;      // "Name" of node. Necessary because index within Node-List may change
                  BoxType           : TBoxType;
                  MiddleX,
                  MiddleY,
                  HalfWidth,
                  HalfHeight        : double;
                  Color             : TColor;
                  Caption           : string;
                  BitmapNodeRect    : TRect;        // Last valid pixel-representation of surrounding rectangle in DotBM
                END;
  TPTDotNode  = ^TDotNode;

  TDotEdge    = RECORD
                  Identifier        : integer;      // "Name" of edge. Necessary because index within edge-List may change
                  SourceIdent,
                  DestIdent         : integer;
                  Caption           : string;
                  CaptionPos        : TPointSgl;
                  CntBPoints        : integer;
                  BPoints           : TBPointArr;
                  BitmapTextRect    : TRect;
                  BitmapArrowRect   : TRect;
                END;
  TPTDotEdge  = ^TDotEdge;
  TArrowEdges = ARRAY[0..2] OF TPoint;

  TDot        = class
                  private
                    DotBM, DotBMEx  : TBitmap;
                    ListDotNode     : TList;
                    ListDotEdge     : TList;
                    GraphWidth,                // Original width of graph in inch
                    GraphHeight     : double;  // Original height of graph in inch
                    SearchEdgeIndex : integer;
                    SearchEdgeIdent : integer;
                    fRunDotActive   : boolean;
                    fWidth,                    // Width of Output-Bitmap in pixels (incl. 2*fBorder)
                    fHeight         : integer; // Height of Output-Bitmap in pixels (incl. 2*fBorder)
                    fBorder         : integer; // Borderwidth in pixels
                    fZoom           : double;  // Zoomfactor
                    fBackCol        : TColor;
                    fNodeHighlight  : integer;
                    fEdgeHighlight  : integer;
                    fPixelsPerInch  : double;  // Graphwidth * fPixelPerInch = Standard Imagewidth (usually 96)
                    GraphDirty,
                    DotBMDirty,
                    DotBMExDirty    : boolean;
                    fDotBinPath     : string;
                    fExecuteWaitEvent: TExecuteWaitEvent;
                    fMarkReversedArrow: boolean;
                    function    GetNextFreeIdentifier: integer;
                    function    GetNodeIndex(Identifier: integer): integer;
                    function    GetEdgeIndex(Identifier: integer): integer;
                    procedure   DeleteNodeByIndex(Index: integer);
                    procedure   DeleteEdgeByIndex(Index: integer);
                    function    RunDot: integer;
                    procedure   ParseDotExportStrings(SL: TStrings);
                    procedure   ParseDotExportStringList(VAR SL: TStringList);
                    procedure   ParseDotExport(FN: string);
                    procedure   CalcArrowEdges(const P1, P2: TPointSgl; VAR P: TArrowEdges);
                    procedure   SetWidth(NewWidth: integer);
                    procedure   SetHeight(NewHeight: integer);
                    procedure   SetZoom(Zoom: double);
                    procedure   SetNodeHighlight(NodeIdent: integer);
                    procedure   SetEdgeHighlight(EdgeIdent: integer);
                    function    AddNodeEx(BoxType: TBoxType; Color: TColor; Caption: string; Identifier: integer): integer;
                    function    AddEdgeEx(IdSource, IdDest: integer; Caption: string; CaptionPos: TPointSgl; CntBPoints: integer; const BPointArr: TBPointArr; Identifier: integer): integer;
                    function    GetNodeCenterByIndex(Index: integer; VAR X, Y: integer): boolean;
                    procedure   PaintArrow(EdgeIndex: integer; Canvas: TCanvas; GraphDestRect: TRect);
                    procedure   PaintNode(NodeIndex: integer; Canvas: TCanvas; GraphDestRect: TRect; HighlightMode, StoreScreenPositions: boolean);
                    procedure   SetMarkReversedArrow(MRA: boolean);
                  public
                    property    MarkReversedArrow: boolean read fMarkReversedArrow write SetMarkReversedArrow;
                    property    DotPath: string read fDotBinPath write fDotBinPath;
                    property    Width: integer read fWidth write SetWidth;
                    property    Height: integer read fHeight write SetHeight;
                    property    Zoom: double read fZoom write SetZoom;
                    property    SelNode: integer read fNodeHighlight write SetNodeHighlight;
                    property    SelEdge: integer read fEdgeHighlight write SetEdgeHighlight;
                    property    ExecuteWaitEvent: TExecuteWaitEvent write fExecuteWaitEvent;
                    constructor Create; virtual;
                    destructor  Destroy; override;
                    procedure   Clear;
                    procedure   Refresh;
                    procedure   SaveToFile(FN: string);
                    procedure   ReadFromFile(const FN: string);
                    function    AddNode(BoxType: TBoxType; Color: TColor; Caption: string): integer;
                    function    AddEdge(IdSource, IdDest: integer; Caption: string): integer;
                    procedure   DeleteNodeByIdentifier(Identifier: integer);
                    procedure   DeleteEdgeByIdentifier(Identifier: integer);
                    function    GetNodeCenterByIdentifier(Identifier: integer; VAR X, Y: integer): boolean;
                    function    GetNearestNodeIdentifier(NodeIdentifier: integer): integer;
                    function    GetNearestNodeIdentifierInDirection(NodeIdentifier: integer; SD: TSearchDirection): integer;
                    function    FindNodeAtPixelXY(X, Y: integer): integer;
                    function    FindEdgeAtPixelXY(X, Y: integer): integer;
                    procedure   GetNodeRect(Identifier: integer; VAR NodeRect: TRect);
                    procedure   GetNodeGraphPos(Identifier: integer; VAR X, Y, W, H: double);
                    function    GetNodeProperties(Identifier: integer; VAR BoxType: TBoxType; VAR Color: TColor; VAR Caption: string): boolean;
                    function    SetNodeProperties(Identifier: integer; BoxType: TBoxType; Color: TColor; Caption: string): boolean;
                    function    GetNodeCaption(Identifier: integer): string;
                    procedure   GetBitmap(VAR BM: TBitmap);
                    function    FindFirstEdge(Identifier: integer; VAR DE: TDotEdge): integer;
                    function    FindNextEdge(VAR DE: TDotEdge): integer;
                    function    GetNodeCount: integer;
                    function    GetNodeIdentifierByIndex(ListIndex: integer; VAR Identifier: integer): boolean;
                    function    GetNodeIdentifierByCaption(Caption: string; VAR Identifier: integer): boolean;
                    procedure   GetEdgePoint(const DotEdge: TDotEdge; PIndex: integer; VAR X, Y: integer);
                    procedure   GetEdgePointEx(Identifier, PIndex: integer; VAR X, Y: integer);
                    procedure   PaintGraphToCanvas(Canvas: TCanvas; GraphDestRect: TRect; StoreScreenPositions: boolean);
                    procedure   PaintGraphToBitmap(VAR BM: TBitmap);
                    procedure   RefreshDotBM;
                    procedure   CalculateBitmapEx;
                    function    GetDotVersion: string;
                    procedure   ExportToWMF(FileName: string);
                    procedure   CopyToClipBoard;
                    procedure   GetNativeGraphDimension(VAR X, Y: double);
                    function    IsDirty: boolean;
                    function    IsRunning: boolean;
                    function    DotAvailable: boolean;
                END;


implementation


constructor TDot.Create;
begin
  inherited Create;
  fRunDotActive     := false;
  GraphWidth        :=  -1;
  GraphHeight       :=  -1;
  ListDotNode       := TList.Create;
  ListDotEdge       := TList.Create;
  fWidth            := 240;
  fHeight           := 320;
  fBorder           :=   7;
  fBackCol          := clWhite;
//  fBackCol := $00E0FFE0;
  fNodeHighlight    :=  -1;
  fEdgeHighlight    :=  -1;
  fPixelsPerInch    :=  96;
  DotBM             := TBitmap.Create;
  DotBM.PixelFormat := pf8Bit;  // Saves a third of memory compared to pf24bit
  DotBM.Width       := fWidth;
  DotBM.Height      := fHeight;
  DotBMEx           := TBitmap.Create;
  DotBMEx.Assign(DotBM);
  GraphDirty        := false;
  DotBMDirty        := false;
  DotBMExDirty      := false;
  // Check default paths for dot.exe
  fDotBinPath       := IncludeTrailingPathDelimiter(ExtractFilePath(ParamStr(0)))+'dot\dot.exe';
  IF not(FileExists(fDotBinPath)) THEN
    fDotBinPath := 'C:\Programme\ATT\Graphviz\bin\dot.exe';
  IF not(FileExists(fDotBinPath)) THEN
    fDotBinPath := 'C:\Program Files\ATT\Graphviz\bin\dot.exe';
  IF not(FileExists(fDotBinPath)) THEN
    fDotBinPath := '';
  MarkReversedArrow:= false;
  fExecuteWaitEvent := nil;
end; {Create}


destructor TDot.Destroy;
begin
  ListDotEdge.Free;
  DotBM.Free;
  Clear;
  ListDotNode.Free;

  inherited Destroy;
end; {Destroy}



FUNCTION ExtractFirstToken(VAR st: string): string;
{ - Returns the first sub-string from st and shortens st. Substrings are separated ba space, except if within "" }
VAR
  i : integer;
BEGIN
  st := Trim(st);
  Result := '';
  IF Length(st) < 1 then exit;

  IF st[1] = '"' then
    begin
      delete(st, 1, 1);
      i := Pos('"', st+'"');
      Result := copy(st, 1, i-1);
      delete(st, 1, i);
    end
    else begin
      i := Pos(' ', st+' ');
      Result := copy(st, 1, i-1);
      delete(st, 1, i);
    end;
END; {ExtractFirstToken}



procedure TDot.SetWidth(NewWidth: integer);
{ - Adjusts Width and Height of DotBM according to Width-Setting }
begin
  IF (GraphWidth <= 0) or (GraphHeight <= 0)
    THEN EXIT;
  IF (NewWidth > 0) and (NewWidth < 32768)
    then begin
      if fWidth <> NewWidth then
        begin
          fWidth  := NewWidth;
          fHeight := 2*fBorder + round((fWidth-2*fBorder) * GraphHeight/GraphWidth);
          fZoom   := (fWidth-2*fBorder) / (GraphWidth * CPixPerInch);
          DotBMDirty := true;
          DotBMExDirty := true;
        end;
    end;
end; {SetWidth}


procedure TDot.SetMarkReversedArrow(MRA: boolean);
begin
  fMarkReversedArrow := MRA;
  DotBMDirty := true;
  DotBMExDirty := true;
end;


procedure TDot.SetHeight(NewHeight: integer);
{ - Adjusts Width and Height of DotBM according to Height-Setting }
begin
  IF (GraphWidth <= 0) or (GraphHeight <= 0)
    THEN EXIT;
  IF (NewHeight > 0) and (NewHeight < 32768)
    then begin
      if fHeight <> NewHeight then
        begin
          fHeight := NewHeight;
          fWidth  := 2*fBorder + round((fHeight-2*fBorder) * GraphWidth/GraphHeight);
          fZoom   := (fHeight-2*fBorder) / (GraphHeight * CPixPerInch);
          DotBMDirty := true;
          DotBMExDirty := true;
        end;
    end;
end; {SetHeight}


procedure TDot.SetZoom(Zoom: double);
begin
  IF GraphHeight > GraphWidth
    then SetHeight(round((GraphHeight * CPixPerInch) * Zoom)+2*fBorder)
    else SetWidth(round((GraphWidth * CPixPerInch) * Zoom)+2*fBorder);
end;


procedure TDot.SetNodeHighlight(NodeIdent: integer);
begin
  IF NodeIdent <> fNodeHighlight then
    begin
      fNodeHighlight := NodeIdent;
      DotBMExDirty := true;
    end;
end;


procedure TDot.SetEdgeHighlight(EdgeIdent: integer);
begin
  IF EdgeIdent <> fEdgeHighlight then
    begin
      fEdgeHighlight := EdgeIdent;
      DotBMExDirty := true;
    end;
end;



procedure TDot.ParseDotExportStrings(SL: TStrings);
{ - Parses dot output and updates Lists
    Excerpt from a dot output file:
      graph 1.000 9.278 21.194
      node n3  8.250 20.944 1.139 0.500 "Main Menu" filled box black #D8FEDB
      edge n25 n28 7 1.875 9.306 1.708 9.222 1.528 9.097 1.375 8.972 1.264 8.861 1.153 8.694 1.069 8.542 F10 1.583 8.861 solid black
      stop
}
var
  SubStr    : ARRAY OF string;
  SubCnt    : integer;
  i, j, k,
  Identifier: integer;
  s         : string;
  CntBPoints: integer;
  BPointArr : TBPointArr;
  CapPos    : TPointSgl;
  BTLoop,
  BoxType   : TBoxType;


        FUNCTION GetDouble(st: string): double;
        VAR
          i : integer;
        BEGIN
          Val(Trim(st), Result, i);
          IF i <> 0 then
            Result := -1;
        END;


        FUNCTION GetIdentifier(s: string): integer;
        { - Returns integer value from string "nxxxx" with x a numerical value }
        VAR
          k : integer;
        BEGIN
          Val(copy(s, 2, Length(s)-1), Result, k);
          IF k <> 0 then
           Result := -1;
        END;


        FUNCTION GetColor(s: string): TColor;
        VAR
          c, k : integer;
        BEGIN
          Result := clWhite;
          IF length(s) <> 7 then
            exit;
          Val('$' + copy(s, 6, 2) + copy(s, 4, 2) + copy(s, 2, 2), c, k);
          IF k = 0 then
            Result := c;
        END;


begin
  Clear;
  FOR j := 0 TO SL.Count-1 DO
    BEGIN
      s := Trim(SL.Strings[j]);
      SubCnt := 0;
      SetLength(SubStr, SubCnt);
      WHILE Length(s) > 0 DO
        BEGIN
          Inc(SubCnt);
          SetLength(SubStr, SubCnt);
          SubStr[SubCnt-1] := ExtractFirstToken(s);
        END;
      IF SubCnt > 0 then
        begin
          SubStr[0] := UpperCase(SubStr[0]);
          IF SubStr[0] = 'GRAPH' THEN
            BEGIN
              GraphWidth  := GetDouble(SubStr[2]);
              GraphHeight := GetDouble(SubStr[3]);
            END;
          IF SubStr[0] = 'NODE' THEN
            BEGIN
              Identifier := GetIdentifier(SubStr[1]);
              IF Identifier >= 0 then
                begin
                  BoxType :=BT_Box;
                  FOR BTLoop := Low(TBoxType) TO High(TBoxType) DO
                    IF UpperCase(SubStr[8]) = UpperCase(CBoxTypeName[BTLoop]) then
                      begin
                        BoxType := BTLoop;
                        break;
                      end;
                  // Dot.exe since version 2.16 seems to add a <> bracket pair to the name
                  IF (Length(SubStr[6]) >= 2) and (SubStr[6][1] = '<') and (SubStr[6][Length(SubStr[6])] = '>') then
                    SubStr[6] := copy(SubStr[6], 2, Length(SubStr[6])-2);
                  i := AddNodeEx(BoxType, GetColor(SubStr[10]), SubStr[6], Identifier);
                  IF i >= 0 then
                    i := ListDotNode.Count-1;
                  IF i >= 0 then
                    WITH TPTDotNode(ListDotNode.Items[i])^  DO
                      begin
                        MiddleX    := GetDouble(SubStr[2]);
                        MiddleY    := GetDouble(SubStr[3]);
                        HalfWidth  := GetDouble(SubStr[4]) * 0.5;
                        HalfHeight := GetDouble(SubStr[5]) * 0.5;
                      end;
                end;
            END;
          IF SubStr[0] = 'EDGE' THEN
            BEGIN
              IF Length(SubStr) < 4
                then CntBPoints := 0
                else CntBPoints := GetIdentifier('n'+SubStr[3]);  // Number of B-Spline Dings
              IF CntBPoints >= 1 then
                begin
                  FOR k := 0 TO CntBPoints-1 DO
                    begin
                      BPointArr[k].X := GetDouble(SubStr[4 + k*2]);
                      BPointArr[k].Y := GetDouble(SubStr[5 + k*2]);
                    end;
                  IF SubCnt > 6+CntBPoints*2+2 then
                    begin
                      Identifier := integer(GetColor(SubStr[5 + CntBPoints*2 + 2]));  // Identifier hidden in Color!
                      s          := SubStr[4 + CntBPoints*2];
                      CapPos.X   := GetDouble(SubStr[4 + CntBPoints*2 + 1]);
                      CapPos.Y   := GetDouble(SubStr[5 + CntBPoints*2 + 1]);
                    end
                    else begin
                      Identifier := integer(GetColor(SubStr[5 + CntBPoints*2]));     // Identifier hidden in Color!
                      s          := '';
                      CapPos.X   := -1;
                      CapPos.Y   := -1;
                    end;
                   IF Identifier = $FFFFFF then  // Standard clWhite in file? }
                     Identifier := GetNextFreeIdentifier;
                   AddEdgeEx(GetIdentifier(SubStr[1]), GetIdentifier(SubStr[2]), s, CapPos, CntBPoints, BPointArr, Identifier);
                end;
            END;
          IF SubStr[0] = 'STOP' THEN
            break;
        end;
    end;
end; {ParseDotExportStrings}






procedure TDot.ParseDotExportStringList(VAR SL: TStringList);
{ - Parses dot output stored in stringlist SL and updates Lists }
var
  j     : integer;
begin
  FOR j := SL.Count-1 DOWNTO 1 DO
    IF Pos('\n', SL.Strings[j]) = 1 then
      begin
        SL.Strings[j-1] := SL.Strings[j-1] + SL.Strings[j];
        SL.Delete(j);
      end;
  ParseDotExportStrings(SL);
end; {ParseDotExportStringList}




procedure TDot.ParseDotExport(FN: string);
{ - Parses dot output in file FN and updates Lists }
var
  SL    : TStringList;
begin
  try
    SL := TStringList.Create;
    SL.LoadFromFile(FN);
    ParseDotExportStringList(SL);
  finally
    SL.Free;
  end;
end; {ParseDotExport}




procedure TDot.SaveToFile(FN: string);
begin
  RunDot;
  copyfile(PChar(IncludeTrailingPathDelimiter(ExtractFilePath(ParamStr(0)))+'temp.txt'), PChar(FN), false);
end;



procedure TDot.ReadFromFile(const FN: string);
begin
  ParseDotExport(FN);
  RunDot;
end;




function TDot.RunDot: integer;
{ - Create a dot language file out of node and edge definitions and calls dot.exe }
var
  i, j     : integer;
  t        : textfile;
  st,
  st1, st2 : string;
  Col      : TColor;
  colbyte  : array[0..3] of byte absolute Col;
begin
  Result := -1;
  if fRunDotActive then
    exit;
  IF not DotAvailable then
    EXIT;

  fRunDotActive := true;
  FileMode := 2;
  assignfile(t, IncludeTrailingPathDelimiter(ExtractFilePath(ParamStr(0)))+'temp.dot');
  rewrite(t);
  writeln(t, 'digraph L0 {');
//  writeln(t, 'Graph G {');
  writeln(t, '  dpi=' + IntToStr(CPixPerInch));
//  writeln(t, '	ordering=out;');
 // writeln(t, '  concentrate=true');
  writeln(t, '  ranksep=1.5');
//  writeln(t, '  aspectratio=1.33');
  writeln(t, '	node [shape = box];');
  writeln(t, '	node [style = filled];');
  writeln(t, '');

  // --- Write nodes to .dot file ----
  for i := 0 to ListDotNode.Count-1 DO
    WITH TPTDotNode(ListDotNode.Items[i])^ DO
      BEGIN
        st1 := ',shape='+CBoxTypeName[BoxType];    // Doppelrahmen: +',peripheries=2';
        j   := integer(Color);
        j   := (j and $FF00) or ((j and $FF0000) shr 16) or ((j and $FF) shl 16);
        st1 := st1 + ',fillcolor="#'+IntToHex(j, 6)+'"';
        st2 := Caption;
        st2 := AnsiReplaceText(st2, #10, '\n');
        writeln(t, '    n'+inttostr(Identifier)+' [label="'+st2+'"'+st1+'];');
      END;

  // --- Write Connections to .dot file ----
  for i := 0 to ListDotEdge.Count-1 DO
    WITH TPTDotEdge(ListDotEdge.Items[i])^ DO
      BEGIN
        IF Caption = ''
          then st := ''
          else st := ' [label="'+Caption+'"]';
        st := st + ' [color="#'+IntToHex(Identifier, 6)+'"]';  // Hide identifier in Color
        writeln(t, '    {n'+inttostr(SourceIdent)+'} -> {n'+inttostr(DestIdent)+' }' +st + ';');
      //' [constraint=false]';
      END;
  writeln(t, '}');
  closefile(t);
//  ExecuteFile(DotBinPath, 'temp.dot -Tjpg -otemp.jpg -Tplain-ext -otemp.txt', ExtractFilePath(ParamStr(0)), true, nil);
  Result := ExecuteFile(fDotBinPath, 'temp.dot -Tplain-ext -otemp.txt', ExtractFilePath(ParamStr(0)), true, fExecuteWaitEvent);
  IF result = 0 then
    begin
      ParseDotExport(IncludeTrailingPathDelimiter(ExtractFilePath(ParamStr(0)))+'temp.txt');
      GraphDirty := false;
      i := fHeight;
      fHeight := 0;
      SetHeight(i); // Adjust Width of DotBM to new Graph
      fEdgeHighlight := -1;
      RefreshDotBM;
    end;
  fRunDotActive := false;
end; {RunDot}



function TDot.GetNextFreeIdentifier: integer;
{ - Returns the next free integer value for naming a node or an edge }
var
  i : integer;
begin
  Result := 0;
  FOR i := ListDotEdge.Count-1 DOWNTO 0 DO
    WITH TPTDotEdge(ListDotEdge.Items[i])^ DO
      IF Result <= Identifier THEN
        Result := Identifier+1;
  FOR i := ListDotNode.Count-1 DOWNTO 0 DO
    WITH TPTDotNode(ListDotNode.Items[i])^ DO
      IF Result <= Identifier THEN
        Result := Identifier+1;
end; {GetNextFreeIdentifier}



function TDot.GetNodeIndex(Identifier: integer): integer;
{ - Returns ListDotNode Index of the node with the given identifier.
    Returns -1 if node doesn't exist }
var
  i : integer;
begin
  Result := -1;
  IF Identifier >= 0 then
    FOR i := 0 TO ListDotNode.Count-1 DO
      if TPTDotNode(ListDotNode.Items[i])^.Identifier = Identifier then
        begin
          Result := i;
          break;
        end;
end; {GetNodeIndex}



function TDot.GetEdgeIndex(Identifier: integer): integer;
{ - Returns ListDotEdge Index of the edge with the given identifier.
    Returns -1 if node doesn't exist }
var
  i : integer;
begin
  Result := -1;
  IF Identifier >= 0 then
    FOR i := 0 TO ListDotEdge.Count-1 DO
      if TPTDotEdge(ListDotEdge.Items[i])^.Identifier = Identifier then
        begin
          Result := i;
          break;
        end;
end; {GetEdgeIndex}



procedure TDot.Clear;
var
  i : integer;
begin
  for i := ListDotNode.Count-1 DOWNTO 0 DO
    DeleteNodeByIndex(i);
  for i := ListDotEdge.Count-1 DOWNTO 0 DO
    DeleteEdgeByIndex(i);
  GraphDirty   := true;
  DotBMDirty   := true;
  DotBMExDirty := true;
end;



procedure TDot.DeleteNodeByIndex(Index: integer);
VAR
  i,
  NodeIdent : integer;
  p         : TPTDotNode;
begin
  if (Index<0) or (Index >= ListDotNode.Count) then
    exit;

  // Remove all edges pointing to that node
  NodeIdent := TPTDotNode(ListDotNode.Items[Index])^.Identifier;
  FOR i := ListDotEdge.Count-1 DOWNTO 0 DO
    WITH TPTDotEdge(ListDotEdge.Items[i])^ DO
      IF (SourceIdent = NodeIdent) or (DestIdent = NodeIdent) THEN
        DeleteEdgeByIndex(i);

  // Free node memory and delete node
  TPTDotNode(ListDotNode.Items[Index])^.Caption := '';
  p := TPTDotNode(ListDotNode.Items[Index]);
  Dispose(p);
  ListDotNode.Delete(Index);

  GraphDirty   := true;
  DotBMDirty   := true;
  DotBMExDirty := true;
end; {DeleteNodeByIndex}



procedure TDot.DeleteEdgeByIndex(Index: integer);
{ - Deletes edge "Index" from ListDotEdge. Doesn't refresh afterwards! }
VAR
  p : TPTDotEdge;
begin
  if (Index<0) or (Index >= ListDotEdge.Count) then
    exit;

  // Free edge memory and delete edge
  TPTDotEdge(ListDotEdge.Items[Index])^.Caption := '';
  p := TPTDotEdge(ListDotEdge.Items[Index]);
  Dispose(p);
  ListDotEdge.Delete(Index);

  fEdgeHighlight := -1;
  GraphDirty   := true;
  DotBMDirty   := true;
  DotBMExDirty := true;
end; {DeleteEdgeByIndex}



procedure TDot.DeleteEdgeByIdentifier(Identifier: integer);
begin
  DeleteEdgeByIndex(GetEdgeIndex(Identifier));
  Refresh;
end;



function TDot.AddNodeEx(BoxType: TBoxType; Color: TColor; Caption: string; Identifier: integer): integer;
{ - Adds a new node. Returns Identifier if sucessfull, -1 else }
var
  PDotNode: TPTDotNode;
begin
  New(PDotNode);
  PDotNode^.Identifier := Identifier;
  PDotNode^.BoxType    := BoxType;
  PDotNode^.Color      := Color;
  PDotNode^.Caption    := Caption;
  PDotNode^.HalfWidth  := -1;
  PDotNode^.HalfHeight := -1;
  ListDotNode.Add(PDotNode);
  GraphDirty   := true;
  DotBMDirty   := true;
  DotBMExDirty := true;
  Result := PDotNode^.Identifier;
end;



function TDot.AddNode(BoxType: TBoxType; Color: TColor; Caption: string): integer;
{ - Adds a new node. Returns identifier given to that new node }
begin
  Result := AddNodeEx(BoxType, Color, Caption, GetNextFreeIdentifier);
end; {AddNode}




procedure TDot.DeleteNodeByIdentifier(Identifier: integer);
{ - Deletes a node and all its connections }
begin
  // Delete node
  DeleteNodeByIndex(GetNodeIndex(Identifier));

  GraphDirty   := true;
  DotBMDirty   := true;
  DotBMExDirty := true;
  Refresh;
end; {DeleteNode}



function TDot.AddEdgeEx(IdSource, IdDest: integer; Caption: string; CaptionPos: TPointSgl; CntBPoints: integer; const BPointArr: TBPointArr; Identifier: integer): integer;
{ - Adds a new edge. Returns identifier given to that new edge }
var
  PDotEdge: TPTDotEdge;
begin
  Result := -1;

  IF GetNodeIndex(IdSource) < 0 then
    EXIT;
  IF GetNodeIndex(IdDest) < 0 then
    EXIT;

  New(PDotEdge);
  PDotEdge^.Identifier  := Identifier;
  PDotEdge^.SourceIdent := IdSource;
  PDotEdge^.DestIdent   := IdDest;
  PDotEdge^.Caption     := Caption;
  PDotEdge^.CaptionPos  := CaptionPos;
  IF CntBPoints >= 4 then
    begin
      PDotEdge^.CntBPoints := CntBPoints;
      PDotEdge^.BPoints    := BPointArr;
    end
    else PDotEdge^.CntBPoints := 0;
  ListDotEdge.Add(PDotEdge);
  Result := PDotEdge^.Identifier;

  GraphDirty   := true;
  DotBMDirty   := true;
  DotBMExDirty := true;
end; {AddEdgeEx}




function TDot.AddEdge(IdSource, IdDest: integer; Caption: string): integer;
{ - Adds a new edge. Returns identifier given to that new edge. To get valid
    coordinates for that edge, call RunDot }
var
  DummyPos : TPointSgl;
begin
  DummyPos.X := 0;
  DummyPos.Y := 0;
  Result := AddEdgeEx(IdSource, IdDest, Caption, DummyPos, 0, TPTBPointArr(nil)^, GetNextFreeidentifier);
end; {AddEdge}




function TDot.GetNodeCenterByIndex(Index: integer; VAR X, Y: integer): boolean;
begin
  Result := false;
  X := -1;
  Y := -1;
  IF (GraphWidth < 0) or (GraphHeight < 0) or
     (DotBM.Width <= 0) or (DotBM.Height <= 0) then
    EXIT;
  Refresh;
  Result := (Index >= 0) and (Index < ListDotNode.Count);
  if Result then
    WITH TPTDotNode(ListDotNode.Items[Index])^  DO
      begin
        X := trunc(MiddleX * DotBM.Width / GraphWidth);
        Y := trunc(DotBM.Height - MiddleY * DotBM.Height / GraphHeight);
      end;
end; {GetNodeCenterByIndex}


function TDot.GetNodeCenterByIdentifier(Identifier: integer; VAR X, Y: integer): boolean;
begin
  Result := GetNodeCenterByIndex(GetNodeIndex(Identifier), X, Y);
end; {GetNodeCenterByIdentifier}



function TDot.GetNearestNodeIdentifier(NodeIdentifier: integer): integer;
{ - Returns indentifier of nearest node relative to node "NodeIdentifier" }
var
  i, X, Y, Index,
  LastDist, d,
  NodeCount,
  X0, Y0      : integer;
begin
  Result := -1;
  LastDist := MaxLongint;
  Index := GetNodeIndex(NodeIdentifier);
  if not GetNodeCenterByIndex(Index, X0, Y0) then
    exit;
  NodeCount := GetNodeCount;
  FOR i := 0 TO NodeCount-1 DO
    IF i <> Index then
      begin
        GetNodeCenterByIndex(i, X, Y);
        X := X-X0;
        Y := Y-Y0;
        d := X*X + Y*Y;
        if d < LastDist then
          begin
            Result := i;
            LastDist := d;
          end;
      end;
end; {GetNearestNodeIdentifier}




function TDot.GetNearestNodeIdentifierInDirection(NodeIdentifier: integer; SD: TSearchDirection): integer;
{ - Searches next node in graph relative to the selected one depending on given search direction }
var
  i, Index,
  X0, Y0,
  X, Y, d,
  NodeCount : integer;
  NextId    : integer;
  LastDist  : integer;
  Valid     : boolean;
begin
  Result := -1;
  IF GetNodeCount < 2 then
    exit;

  Index := GetNodeIndex(NodeIdentifier);
  NextID := -1;
  LastDist := MaxLongint;
  GetNodeCenterByIndex(Index, X0, Y0);
  NodeCount := GetNodeCount;
  FOR i := 0 TO NodeCount-1 DO
    IF i <> Index then
      begin
        GetNodeCenterByIndex(i, X, Y);
        X := (X-X0);
        Y := (Y-Y0);
        CASE SD OF
          SDAbove: Valid := (Y < 0) and (Abs(X)-Abs(Y) < 0);
          SDRight: Valid := (X > 0) and (Abs(X)-Abs(Y) > 0);
          SDBelow: Valid := (Y > 0) and (Abs(X)-Abs(Y) < 0);
          SDLeft : Valid := (X < 0) and (Abs(X)-Abs(Y) > 0);
          ELSE Valid := false;
        END; {case}
        IF Valid then
          begin
            d := X*X + Y*Y;
            IF d < LastDist then
              begin
                LastDist := d;
                NextID := i;
              end
          end;
      end;
  IF NextID >= 0
    then Result := TPTDotNode(ListDotNode.Items[NextID])^.Identifier
    else Result := -1;
end;





function TDot.FindNodeAtPixelXY(X, Y: integer): integer;
{ - Returns identifier of node at given position }
var
  i : integer;
  P : TPoint;
begin
  Result := -1;
  Refresh;  
  IF (GraphWidth < 0) or (GraphHeight < 0) or
     (DotBM.Width <= 0) or (DotBM.Height <= 0) then
    EXIT;
  P.X := X;
  P.Y := Y;
  FOR i := 0 TO ListDotNode.Count-1 DO
    WITH TPTDotNode(ListDotNode.Items[i])^ DO
      IF PtInRect(Rect(BitmapNodeRect.Left-3,
                       BitmapNodeRect.Top-3,
                       BitmapNodeRect.Right+3,
                       BitmapNodeRect.Bottom+3), P) then
        BEGIN
          Result := Identifier;
          break;
        END;
end; {FindNodeAtXY}


function TDot.FindEdgeAtPixelXY(X, Y: integer): integer;
{ - Returns identifier of edge at given position (Label and Arrow are hotspots) }
VAR
  i, e : integer;
  P    : TPoint;
begin
  P.X := X;
  P.Y := Y;
  Result := -1;
  Refresh;
  IF (GraphWidth < 0) or (GraphHeight < 0) or
     (DotBM.Width <= 0) or (DotBM.Height <= 0) then
    EXIT;

  IF fEdgeHighlight >= 0
    then e := GetEdgeIndex(fEdgeHighlight)+1
    else e := 0;

  FOR i := 0 TO ListDotEdge.Count-1 DO
    WITH TPTDotEdge(ListDotEdge.Items[(i+e) mod ListDotEdge.Count])^ DO
      if PtInRect(BitmapTextRect, P) then
        begin
          Result := Identifier;
          break;
        end;
  IF Result = -1 THEN
    FOR i := 0 TO ListDotEdge.Count-1 DO
      WITH TPTDotEdge(ListDotEdge.Items[(i+e) mod ListDotEdge.Count])^ DO
        if PtInRect(BitmapArrowRect, P) then
          begin
            Result := Identifier;
            break;
          end;
end; {FindEdgeAtPixelXY}



procedure TDot.Refresh;
begin
  IF GraphDirty THEN
    RunDot;
  IF GraphDirty THEN // Still dirty?
    EXIT;
  IF DotBMDirty THEN
    RefreshDotBM;
  IF DotBMExDirty THEN
    CalculateBitmapEx;
end; {Refresh}



procedure TDot.GetBitmap(VAR BM: TBitmap);
begin
  Refresh;
  BM.Assign(DotBMEx);
end; {GetBitmap}



procedure TDot.GetNodeRect(Identifier: integer; VAR NodeRect: TRect);
VAR
  W, H : integer;
begin
  FillChar(NodeRect, sizeOf(NodeRect), #0);
  Refresh;
  Identifier := GetNodeIndex(Identifier);
  IF (Identifier < 0) or (GraphWidth <= 0) or (GraphHeight <= 0) then
    exit;
  W := DotBM.Width-16;
  H := DotBM.Height-16;
  WITH TPTDotNode(ListDotNode.Items[Identifier])^  DO
    NodeRect := Rect(8+round((MiddleX-HalfWidth)/GraphWidth*W),
                     DotBM.Height-8-round((MiddleY-HalfHeight)/GraphHeight*H),
                     8+round((MiddleX+HalfWidth)/GraphWidth*W),
                     DotBM.Height-8-round((MiddleY+HalfHeight)/GraphHeight*H));
end; {GetNodeRect}



procedure TDot.GetNodeGraphPos(Identifier: integer; VAR X, Y, W, H: double);
begin
  X := 0;
  Y := 0;
  W := 0;
  H := 0;
  Refresh;
  Identifier := GetNodeIndex(Identifier);
  IF (Identifier < 0) or (GraphWidth <= 0) or (GraphHeight <= 0) then
    exit;
  W := DotBM.Width-16;
  H := DotBM.Height-16;
  WITH TPTDotNode(ListDotNode.Items[Identifier])^  DO
    BEGIN
      X := MiddleX;
      Y := MiddleY;
      W := HalfWidth;
      H := HalfHeight;
    END;
end; {GetNodeGraphPos}



function TDot.GetNodeProperties(Identifier: integer; VAR BoxType: TBoxType; VAR Color: TColor; VAR Caption: string): boolean;
begin
  Identifier := GetNodeIndex(Identifier);
  IF (Identifier < 0) then
    begin
      BoxType := BT_Box;
      Color   := clBlack;
      Caption := '';
      Result  := false;
      exit;
    end;
  BoxType := TPTDotNode(ListDotNode.Items[Identifier])^.BoxType;
  Color   := TPTDotNode(ListDotNode.Items[Identifier])^.Color;
  Caption := TPTDotNode(ListDotNode.Items[Identifier])^.Caption;
  Result  := true;
end; {GetNodeProperties}



function TDot.SetNodeProperties(Identifier: integer; BoxType: TBoxType; Color: TColor; Caption: string): boolean;
begin
  Identifier := GetNodeIndex(Identifier);
  IF (Identifier < 0) then
    begin
      Result  := false;
      exit;
    end;
  GraphDirty := (TPTDotNode(ListDotNode.Items[Identifier])^.BoxType <> BoxType) or
                (TPTDotNode(ListDotNode.Items[Identifier])^.Caption <> Caption);
  DotBMDirty := (TPTDotNode(ListDotNode.Items[Identifier])^.Color   <> Color) or GraphDirty;
  DotBMExDirty := DotBMDirty;

  TPTDotNode(ListDotNode.Items[Identifier])^.BoxType := BoxType;
  TPTDotNode(ListDotNode.Items[Identifier])^.Color   := Color;
  TPTDotNode(ListDotNode.Items[Identifier])^.Caption := Caption;
  Result  := true;
end;


function TDot.GetNodeCaption(Identifier: integer): string;
VAR
  BoxType : TBoxType;
  Color   : TColor;
begin
  GetNodeProperties(Identifier, BoxType, Color, Result);
end;



function TDot.FindFirstEdge(Identifier: integer; VAR DE: TDotEdge): integer;
{ - Searches the first edge that points to or from the specified Node
    Returns 0 if sucessfull }
begin
  SearchEdgeIndex := 0;
  SearchEdgeIdent := Identifier;
  Result := FindNextEdge(DE);
end; {FindFirstEdge}



function TDot.FindNextEdge(VAR DE: TDotEdge): integer;
{ - Searches the next edge that points to or from the specified Node. Call FindFirstEdge first.
    Returns 0 if sucessfull }
begin
  FillChar(DE, SizeOf(DE), #0);
  Result := -1;
  WHILE SearchEdgeIndex < ListDotEdge.Count do
    WITH TPTDotEdge(ListDotEdge.Items[SearchEdgeIndex])^ DO
      begin
        IF (SourceIdent = SearchEdgeIdent) or (DestIdent = SearchEdgeIdent) then
          begin
            DE := TPTDotEdge(ListDotEdge.Items[SearchEdgeIndex])^;
            Result := 0;
            Inc(SearchEdgeIndex);
            break;
          end;
        Inc(SearchEdgeIndex);
      end;
end; {FindNextEdge}



function TDot.GetNodeIdentifierByIndex(ListIndex: integer; VAR Identifier: integer): boolean;
{ - Returns Identifier of node stored in position ListIndex within the TDot-internal Node list }
begin
  Identifier := -1;
  Result := false;
  IF (ListIndex < 0) or (ListIndex >= ListDotNode.Count) then
    exit;
  Identifier := TPTDotNode(ListDotNode.Items[ListIndex])^.Identifier;
  Result := true;
end; {GetNodeidentifierByIndex}


function TDot.GetNodeIdentifierByCaption(Caption: string; VAR Identifier: integer): boolean;
VAR
  i : integer;
begin
  Result := false;
  Identifier := -1;
  Caption := UpperCase(Caption);
  FOR i := 0 TO ListDotNode.Count-1 DO
    IF Caption = UpperCase(TPTDotNode(ListDotNode.Items[i])^.Caption) THEN
      BEGIN
        Identifier := TPTDotNode(ListDotNode.Items[i])^.Identifier;
        Result := true;
        break;
      END;
end;


function TDot.GetNodeCount: integer;
begin
  Result := ListDotNode.Count;
end; {GetNodeCount}


procedure TDot.GetEdgePoint(const DotEdge: TDotEdge; PIndex: integer; VAR X, Y: integer);
begin
  X := 0;
  Y := 0;
  IF (PIndex >= 0) and (PIndex < DotEdge.CntBPoints) then
    begin
      X := trunc(fBorder + DotEdge.BPoints[PIndex].X * (DotBM.Width-16) / GraphWidth);
      Y := trunc(DotBM.Height - fBorder-DotEdge.BPoints[PIndex].Y * (DotBM.Width-16) / GraphWidth);
    end;
end; {GetEdgePoint}


procedure TDot.GetEdgePointEx(Identifier, PIndex: integer; VAR X, Y: integer);
begin
  X := 0;
  Y := 0;
  Identifier := GetEdgeIndex(Identifier);
  IF Identifier >= 0 then
    GetEdgePoint(TPTDotEdge(ListDotEdge[Identifier])^, PIndex, X, Y);
end; {GetEdgePointEx}


procedure TDot.CalcArrowEdges(const P1, P2: TPointSgl; VAR P: TArrowEdges);
{ - Calculates edges of arrow based on the last two points of the arrow's line }
var
  d : double;
begin
  d := Sqrt(Sqr(P1.X-P2.X)+Sqr(P1.Y-P2.Y));
  IF d > 0 then
    d := 0.15 / d;
  P[0].X := round((P1.X + (P1.X-P2.X)*d)*fPixelsPerInch);
  P[0].Y := round((P1.Y + (P1.Y-P2.Y)*d)*fPixelsPerInch);
  d := d / 3;
  P[1].X := round((P1.X + (P1.Y-P2.Y)*d)*fPixelsPerInch);
  P[1].Y := round((P1.Y - (P1.X-P2.X)*d)*fPixelsPerInch);
  P[2].X := round((P1.X - (P1.Y-P2.Y)*d)*fPixelsPerInch);
  P[2].Y := round((P1.Y + (P1.X-P2.X)*d)*fPixelsPerInch);
end; {CalcArrowEdges}



procedure TDot.PaintArrow(EdgeIndex: integer; Canvas: TCanvas; GraphDestRect: TRect);
var
  c, j, X0, Y0 : integer;
  Points       : ARRAY[0..CMaxBPoint] OF TPoint;
  ArrowEdge    : TArrowEdges;
  ColChanged   : boolean;
begin
  IF (EdgeIndex < 0) or (EdgeIndex >= ListDotEdge.Count) THEN
    EXIT;
  ColChanged   := false;
  X0 := GraphDestRect.Left + fBorder;
  Y0 := GraphDestRect.Bottom - fBorder;
  c := TPTDotEdge(ListDotEdge[EdgeIndex])^.CntBPoints;
  IF c >= 4 then
    begin
      FOR j := 0 TO c-1 DO
        begin
          Points[j].X := X0 + trunc(TPTDotEdge(ListDotEdge[EdgeIndex])^.BPoints[j].X * fPixelsPerInch);
          Points[j].Y := Y0 - trunc(TPTDotEdge(ListDotEdge[EdgeIndex])^.BPoints[j].Y * fPixelsPerInch);
        end;
      if MarkReversedArrow and (Canvas.Pen.Color = clBlack)
        then begin
          IF Points[c-1].Y < Points[0].Y THEN
            BEGIn
              Canvas.Pen.Color := clFuchsia;
              Canvas.Brush.Color := clFuchsia;
              Canvas.Pen.Width := 2;
              ColChanged       := true;
            END;
        end;
      Canvas.PolyBezier(Slice(Points, c));

      CalcArrowEdges(TPTDotEdge(ListDotEdge[EdgeIndex])^.BPoints[c-1],
                     TPTDotEdge(ListDotEdge[EdgeIndex])^.BPoints[c-2], ArrowEdge);
      FOR j := 0 TO 2 DO
        begin
          ArrowEdge[j].X := X0 + ArrowEdge[j].X;
          ArrowEdge[j].Y := Y0 - ArrowEdge[j].Y;
        end;
      Canvas.Polygon(ArrowEdge);
      TPTDotEdge(ListDotEdge[EdgeIndex])^.BitmapArrowRect := Rect(Min(Min(ArrowEdge[0].X, ArrowEdge[1].X), ArrowEdge[2].X)-1,
                                                                  Min(Min(ArrowEdge[0].Y, ArrowEdge[1].Y), ArrowEdge[2].Y)-1,
                                                                  Max(Max(ArrowEdge[0].X, ArrowEdge[1].X), ArrowEdge[2].X)+2,
                                                                  Max(Max(ArrowEdge[0].Y, ArrowEdge[1].Y), ArrowEdge[2].Y)+2);
      IF ColChanged then
        begin
          Canvas.Pen.Color := clBlack;
          Canvas.Brush.Color := clBlack;
          Canvas.Pen.Width := 1;
        end;
    end;
end; {PaintArrow}



procedure TDot.PaintNode(NodeIndex: integer; Canvas: TCanvas; GraphDestRect: TRect; HighlightMode, StoreScreenPositions: boolean);
var
  X0, Y0  : integer;
  i, j    : integer;
  Points  : ARRAY[0..CMaxBPoint] OF TPoint;
  SRect   : TRect;


       PROCEDURE line(X0, Y0, X1, Y1: integer);
       BEGIN
         Canvas.MoveTo(X0, Y0);
         Canvas.LineTo(X1, Y1);
       END;

begin
  IF (NodeIndex < 0) or (NodeIndex >= ListDotNode.Count) then
    exit;
  X0 := GraphDestRect.Left + fBorder;
  Y0 := GraphDestRect.Bottom - fBorder;
  WITH TPTDotNode(ListDotNode[NodeIndex])^ DO
    begin
      SRect := Rect(X0 + trunc((MiddleX - HalfWidth )*fPixelsPerInch),
                    Y0 - trunc((MiddleY + HalfHeight)*fPixelsPerInch),
                    X0 + trunc((MiddleX + HalfWidth )*fPixelsPerInch),
                    Y0 - trunc((MiddleY - HalfHeight)*fPixelsPerInch));
      IF StoreScreenPositions THEN
        BitmapNodeRect := SRect;
      CASE Boxtype OF
        BT_None: if HighlightMode then
                   begin
                     Canvas.Pen.Style := psDot;
                     Canvas.Pen.Width := 1;
                     Canvas.Rectangle(SRect);
                     Canvas.Pen.Style := psSolid;
                     Canvas.Rectangle(SRect.Left-1, SRect.Top-1, SRect.Right+1, SRect.Bottom+1);
                   end;
        BT_Box : begin
                   Canvas.Rectangle(SRect);
                 end;
        BT_Diamond: begin
                   Points[0].X := (SRect.Left + SRect.Right) div 2;
                   Points[0].Y :=  SRect.Top;
                   Points[1].X :=  SRect.Right;
                   Points[1].Y := (SRect.Top + SRect.Bottom) div 2;
                   Points[2].X :=  Points[0].X;
                   Points[2].Y :=  SRect.Bottom;
                   Points[3].X :=  SRect.Left;
                   Points[3].Y :=  Points[1].Y;
                   Points[4] := Points[0];
                   Canvas.Polygon(Slice(Points, 5));
                 end;
        BT_Circle,
        BT_Ellipse: begin
                   Canvas.Ellipse(SRect);
                 end;
        BT_Hexagon: begin
                   i := (SRect.Right-SRect.Left) div 4;
                   j := SRect.Right - i;
                   i := SRect.Left + i;
                   Points[0].X :=  i;
                   Points[0].Y :=  SRect.Top;
                   Points[1].X :=  j;
                   Points[1].Y :=  SRect.Top;
                   Points[2].X :=  SRect.Right;
                   Points[2].Y := (SRect.Top + SRect.Bottom) div 2;
                   Points[3].X :=  j;
                   Points[3].Y :=  SRect.Bottom;
                   Points[4].X :=  i;
                   Points[4].Y :=  SRect.Bottom;
                   Points[5].X :=  SRect.Left;
                   Points[5].Y := (SRect.Top + SRect.Bottom) div 2;
                   Points[6] := Points[0];
                   Canvas.Polygon(Slice(Points, 6));
                   end;
        BT_Triangle: begin
                   j := (SRect.Bottom-SRect.Top) div 4;
                   Points[0].X :=  SRect.Left;
                   Points[0].Y :=  SRect.Bottom-j;
                   Points[1].X := (SRect.Left + SRect.Right) div 2;
                   Points[1].Y :=  SRect.Top;
                   Points[2].X :=  SRect.Right;
                   Points[2].Y :=  SRect.Bottom-j;
                   Points[3] := Points[0];
                   Canvas.Polygon(Slice(Points, 4));
                 end;
        BT_InvTriangle: begin
                   j := (SRect.Bottom-SRect.Top) div 4;
                   Points[0].X :=  SRect.Right;
                   Points[0].Y :=  SRect.Top+j;
                   Points[1].X := (SRect.Left + SRect.Right) div 2;
                   Points[1].Y :=  SRect.Bottom;
                   Points[2].X :=  SRect.Left;
                   Points[2].Y :=  SRect.Top+j;
                   Points[3] := Points[0];
                   Canvas.Polygon(Slice(Points, 4));
                 end;
        BT_Trapezium: begin
                   i := (SRect.Right-SRect.Left) div 5;
                   Points[0].X :=  SRect.Left+i;
                   Points[0].Y :=  SRect.Top;
                   Points[1].X :=  SRect.Right-i;
                   Points[1].Y :=  SRect.Top;
                   Points[2].X :=  SRect.Right;
                   Points[2].Y :=  SRect.Bottom;
                   Points[3].X :=  SRect.Left;
                   Points[3].Y :=  SRect.Bottom;
                   Points[4] := Points[0];
                   Canvas.Polygon(Slice(Points, 5));
                 end;
        BT_Parallelogram: begin
                   i := (SRect.Right-SRect.Left) div 5;
                   Points[0].X :=  SRect.Left+i;
                   Points[0].Y :=  SRect.Top;
                   Points[1].X :=  SRect.Right;
                   Points[1].Y :=  SRect.Top;
                   Points[2].X :=  SRect.Right-i;
                   Points[2].Y :=  SRect.Bottom;
                   Points[3].X :=  SRect.Left;
                   Points[3].Y :=  SRect.Bottom;
                   Points[4] := Points[0];
                   Canvas.Polygon(Slice(Points, 5));
                 end;
      END; {case}
    end;
end;



procedure TDot.PaintGraphToCanvas(Canvas: TCanvas; GraphDestRect: TRect; StoreScreenPositions: boolean);
{ - Paints Graph to Canvas within GraphDestRect }
var
  i, j, c,
  w, h,
  x, y,
  X0, Y0    : integer;
  Width,
  Height,
  MaxWidth,
  MaxHeight : integer;
  st, s2    : string;
  SRect     : TRect;
begin
  if (GraphWidth<= 0) or (GraphHeight <= 0) then
    exit;

  WITH GraphDestRect DO
    BEGIN
      Width  := Right-Left+1;
      Height := Bottom-Top+1;
    END;

  MaxWidth     := Width  - 2*fBorder;
  MaxHeight    := Height - 2*fBorder;

  IF GraphWidth > GraphHeight
    then fPixelsPerInch := MaxWidth/GraphWidth
    else fPixelsPerInch := MaxHeight/GraphHeight;

  WITH Canvas DO
    BEGIN
      Pen.Color   := fBackCol;
      Brush.color := fBackCol;
      Brush.style := bsSolid;
      Rectangle(-1, -1, Width+1, height+1);
      Pen.Color   := clBlack;
    END;

  X0 := GraphDestRect.Left + fBorder;
  Y0 := GraphDestRect.Bottom - fBorder;

  // Paint Edges (Arrows).....
  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := clBlack;
  FOR i := 0 TO ListDotEdge.Count-1 DO
    PaintArrow(i, Canvas, GraphDestRect);

  { Canvas.Font.Size := round(12 * fPixelsPerInch/CPixPerInch);  // 12pt is standard in dot.exe }
  Canvas.Font.Size := round(12 * fPixelsPerInch/Canvas.Font.PixelsPerInch);  // 12pt is standard in dot.exe

  // Paint Caption of edges .....
  FOR i := 0 TO ListDotEdge.Count-1 DO
    BEGIN
      c := TPTDotEdge(ListDotEdge[i])^.CntBPoints;
      IF c >= 4 then
        begin
          st := TPTDotEdge(ListDotEdge[i])^.Caption;
          if Trim(st) <> '' then
            begin
              X := X0 + trunc(TPTDotEdge(ListDotEdge[i])^.CaptionPos.X * fPixelsPerInch);
              Y := Y0 - trunc(TPTDotEdge(ListDotEdge[i])^.CaptionPos.Y * fPixelsPerInch);
              WITH Canvas DO
                BEGIN
                  w := TextWidth(st);
                  h := Textheight(st);
                  Brush.Style := bsClear;
                  Font.Color  := clBlack;
                  SRect := Rect(X-w div 2, Y-h div 2, X+w div 2, Y+h div 2);
                  IF StoreScreenPositions THEN
                    TPTDotEdge(ListDotEdge[i])^.BitmapTextRect := SRect;
                  TextOut(SRect.Left, SRect.Top, st);
                END;
            end
            else IF StoreScreenPositions THEN
              TPTDotEdge(ListDotEdge[i])^.BitmapTextRect := Rect(-1, -1, -1, -1);
        end
        else IF StoreScreenPositions THEN
          TPTDotEdge(ListDotEdge[i])^.BitmapTextRect := Rect(-1, -1, -1, -1);
    END;

  // Paint Nodes...
  Canvas.Brush.Style := bsSolid;
  FOR i := 0 TO ListDotNode.Count-1 DO
    WITH TPTDotNode(ListDotNode[i])^ DO
      begin
        Canvas.Brush.Color := Color;
        PaintNode(i, Canvas, GraphDestRect, false, StoreScreenPositions);
      end;
  // Write Node text...
  Canvas.Brush.Style := bsClear;
  Canvas.Font.Color  := clBlack;
  Canvas.Font.Name   := 'Times New Roman';
  FOR i := 0 TO ListDotNode.Count-1 DO
    WITH TPTDotNode(ListDotNode[i])^ DO
      begin
        st := Caption;
        st := AnsiReplaceText(st, '\n', #10);
        // Count lines of text...
        c := 0;
        FOR j := 1 TO Length(st) DO
          IF st[j] = #10 THEN Inc(c);
        IF length(st) > 0
          then Inc(c);
        // Textout...
        st := st + #10;
        X := trunc(X0 + MiddleX*fPixelsPerInch);
        FOR j := 1 TO c DO
          begin
            Y := trunc(Y0 - MiddleY*fPixelsPerInch);
            s2 := copy(st, 1, Pos(#10, st)-1);
            delete(st, 1, Length(s2)+1);
            WITH Canvas DO
              TextOut(X-TextWidth(s2) div 2, Y+ trunc((j-c/2-1)*Textheight(s2)), s2);
          end;
      end;
end;



procedure TDot.PaintGraphToBitmap(VAR BM: TBitmap);
BEGIN
  BM.Width  := fWidth;
  BM.Height := fHeight;
  PaintGraphToCanvas(BM.Canvas, Rect(0, 0, BM.Width, BM.Height), BM = DotBM);
END;


procedure TDot.RefreshDotBM;
begin
  PaintGraphToBitmap(DotBM);
  DotBMDirty := false;
end;



function GetConsoleOutput(const Command: String; var Output, Errors: TStringList): Boolean;
// http://www.dsdt.info/tipps/?id=637
var
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  SecurityAttr: TSecurityAttributes;
  PipeOutputRead: THandle;
  PipeOutputWrite: THandle;
  PipeErrorsRead: THandle;
  PipeErrorsWrite: THandle;
  Succeed: Boolean;
  Buffer: array [0..255] of Char;
  NumberOfBytesRead: DWORD;
  Stream: TMemoryStream;
begin
  //Initialisierung ProcessInfo
  FillChar(ProcessInfo, SizeOf(TProcessInformation), 0);

  //Initialisierung SecurityAttr
  FillChar(SecurityAttr, SizeOf(TSecurityAttributes), 0);
  SecurityAttr.nLength := SizeOf(SecurityAttr);
  SecurityAttr.bInheritHandle := true;
  SecurityAttr.lpSecurityDescriptor := nil;

  //Pipes erzeugen
  CreatePipe(PipeOutputRead, PipeOutputWrite, @SecurityAttr, 0);
  CreatePipe(PipeErrorsRead, PipeErrorsWrite, @SecurityAttr, 0);

  //Initialisierung StartupInfo
  FillChar(StartupInfo, SizeOf(TStartupInfo), 0);
  StartupInfo.cb:=SizeOf(StartupInfo);
  StartupInfo.hStdInput := 0;
  StartupInfo.hStdOutput := PipeOutputWrite;
  StartupInfo.hStdError := PipeErrorsWrite;
  StartupInfo.wShowWindow := sw_Hide;
  StartupInfo.dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;

  if  CreateProcess(nil, PChar(command), nil, nil, true,
  CREATE_DEFAULT_ERROR_MODE or CREATE_NEW_CONSOLE or NORMAL_PRIORITY_CLASS, nil, nil,
  StartupInfo, ProcessInfo) then begin
    result:=true;
    //Write-Pipes schlieen
    CloseHandle(PipeOutputWrite);
    CloseHandle(PipeErrorsWrite);

    //Ausgabe Read-Pipe auslesen
    Stream := TMemoryStream.Create;
    try
      while true do begin
        succeed := ReadFile(PipeOutputRead, Buffer, 255, NumberOfBytesRead, nil);
        if not succeed then break;
        Stream.Write(Buffer, NumberOfBytesRead);
      end;
      Stream.Position := 0;
      Output.LoadFromStream(Stream);
    finally
      Stream.Free;
    end;
    CloseHandle(PipeOutputRead);

    //Fehler Read-Pipe auslesen
    Stream := TMemoryStream.Create;
    try
      while true do begin
        succeed := ReadFile(PipeErrorsRead, Buffer, 255, NumberOfBytesRead, nil);
        if not succeed then break;
        Stream.Write(Buffer, NumberOfBytesRead);
      end;
      Stream.Position := 0;
      Errors.LoadFromStream(Stream);
    finally
      Stream.Free;
    end;
    CloseHandle(PipeErrorsRead);

    WaitForSingleObject(ProcessInfo.hProcess, INFINITE);
    CloseHandle(ProcessInfo.hProcess);
  end
  else begin
    result:=false;
    CloseHandle(PipeOutputRead);
    CloseHandle(PipeOutputWrite);
    CloseHandle(PipeErrorsRead);
    CloseHandle(PipeErrorsWrite);
  end;
end; {GetConsoleOutput}




function TDot.GetDotVersion: string;
VAR
  LN : string;
  var Output, Errors: TStringList;
begin
  Result := '';
  IF (fDotBinPath='') or (not FileExists(fDotBinPath)) then
    EXIT;
  LN := IncludeTrailingPathDelimiter(ExtractFilePath(ParamStr(0))) + 'dotver.log';

  Output := TStringList.Create;
  Errors := TStringList.Create;
  if GetConsoleOutput(fDotBinPath + ' -V', Output, Errors)
    then if Errors.Count > 0 then
      Result := Errors.Strings[0];
  Errors.Free;
  Output.free;
end; {GetDotVersion}




procedure TDot.CalculateBitmapEx;
VAR
  i, EIndex : integer;


        PROCEDURE HighLightArrow(EIndex: integer; Col: TColor);
        BEGIN
          IF (EIndex >= 0) and (EIndex < ListDotEdge.Count) then
            begin
              IF TPTDotEdge(ListDotEdge[EIndex])^.Caption <> '' then
                begin
                  DotBMEx.Canvas.Font.Assign(DotBM.Canvas.Font);
                  DotBMEx.Canvas.Brush.Style := bsClear;
                  DotBMEx.Canvas.Font.Color := Col;
                  DotBMEx.Canvas.TextOut(TPTDotEdge(ListDotEdge[EIndex])^.BitmapTextRect.Left,
                                         TPTDotEdge(ListDotEdge[EIndex])^.BitmapTextRect.Top,
                                         TPTDotEdge(ListDotEdge[EIndex])^.Caption);
                end;
              DotBMEx.Canvas.Pen.Style := psSolid;
              DotBMEx.Canvas.Pen.Color := Col;
              DotBMEx.Canvas.Brush.Style := bsSolid;
              DotBMEx.Canvas.Brush.Color := Col;
              DotBMEx.Canvas.Font.Color := Col;
              PaintArrow(EIndex, DotBMEx.Canvas, Rect(0, 0, DotBMEx.Width, DotBMEx.Height));
            end;
        END;

begin
  DotBMEx.Assign(DotBM);
  DotBMEx.Canvas.Brush.Style := bsSolid;
  DotBMEx.Canvas.Pen.Width := 2;
  IF fNodeHighlight >= 0 then
    begin
      FOR i := 0 TO ListDotNode.Count-1 DO
        WITH TPTDotNode(ListDotNode[i])^ DO
          if fNodeHighLight = Identifier then
            begin
              DotBMEx.Canvas.Brush.style := bsClear;
              DotBMEx.Canvas.Pen.Color := clRed;
              PaintNode(i, DotBMEx.Canvas, Rect(0, 0, DotBMEx.Width, DotBMEx.Height), true, false);
              break;
            end;

      // Highlight all arrows pointing from the selected node
      IF ListDotEdge.Count > 0 then
        begin
          FOR i := 0 TO ListDotEdge.Count-1 DO
            WITH TPTDotEdge(ListDotEdge[i])^ DO
              begin
                IF Sourceident = fNodeHighLight then
                  HighLightArrow(i, clRed);
                IF DestIdent = fNodeHighLight then
                  HighLightArrow(i, $0000C000);
              end;
        end;

    end;

  EIndex := GetEdgeIndex(fEdgeHighlight);
  HighLightArrow(EIndex, clRed);

  DotBMExDirty := false;
end; {CalculateBitmapEx}



FUNCTION GetWMFScaling: double;
{ - There is an unsolved problem with the use of metafiles: I'm not able to
    create a TMetaFileCanvas to paint on that has the correct resolution.
    May have something to do with dpi settings.
    So this routine will calculate a corrective scaling value simple by trying:
    It creates a metafile half black and half white, copies it to a bitmap of
    the same size and than checks where the center really lies. The expected
    scaling value would be 1.0, but in reality it's different. On my laptop the
    value is 0.8478 (about 72/85) }
VAR
  M : TMetafile;
  C : TMetaFileCanvas;
  B : TBitmap;
  i : integer;
BEGIN
  M        := TMetafile.Create;
  M.Inch   := Screen.PixelsPerInch;
  M.Width  := 1024;
  M.Height := 8;

  C             := TMetaFileCanvas.Create(M, 0);
  C.Pen.Style   := psSolid;
  C.Pen.Color   := clBlack;
  C.Brush.Color := clBlack;
  C.Rectangle(0, 0, M.Width div 2, M.Height);
  C.Free;

  Result := 1.0;
  B := TBitmap.Create;
  B.Width := M.Width;
  B.Height:= M.Height;
  B.Canvas.Draw(0, 0, M);
  FOR i := 0 TO B.Width-1 DO     // Slow, I know  ;)
    IF B.Canvas.Pixels[i, 2] > 0 then
      begin
        Result := 2 * i / B.Width;
        break;
      end;
  IF Result > 0 then
    Result := 1 / Result;
  B.Free;
  M.Free;
END;




procedure TDot.ExportToWMF(FileName: string);
{ - Exports graph into WMF file. If no file name is specified, graph is copied to clipboard instead }
VAR
  M        : TMetafile;
  C        : TMetaFileCanvas;
  s        : double;
  AFormat  : word;
  AData    : cardinal;
  APalette : HPalette;
  MMBorder : integer;
begin
  M          := TMetaFile.Create;
  M.enhanced := true;
  {
  M.Inch     := CPixPerInch;
  M.Inch     := Screen.PixelsPerInch;
  }
  M.Inch     := 120;
  MMBorder   := trunc(fBorder/Screen.PixelsPerInch*2540); // Border size conversion from Screen-pixels to 0.01mm units
  M.MMWidth  := round(GraphWidth *2540) + 2*MMBorder;
  M.MMHeight := round(GraphHeight*2540) + 2*MMBorder;

  C := TMetaFileCanvas.CreateWithComment(M, application.MainForm.Canvas.Handle, '', '');

  s := GetWMFScaling;

  PaintGraphToCanvas(C, Rect(0, 0, trunc(M.Width*s)+1, trunc(M.Height*s)+1), false);

  C.Free;
  IF FileName = ''
    then begin
      M.SaveToClipboardFormat(AFormat, AData, APalette);
      ClipBoard.SetAsHandle(AFormat,AData);
    end
    else M.SaveToFile(FileName);
  M.Free;
end;



procedure TDot.CopyToClipBoard;
{ - Copies graph to clipboard }
begin
  ExportToWMF('');
end;


procedure TDot.GetNativeGraphDimension(VAR X, Y: double);
begin
  X := GraphWidth;
  Y := GraphHeight;
end;



function TDot.IsDirty: boolean;
begin
  Result := GraphDirty;
end;


function TDot.IsRunning: boolean;
begin
  Result := fRunDotActive;
end;


function TDot.DotAvailable: boolean;
begin
  Result := (fDotBinPath <> '') and (FileExists(fDotBinPath));
end;



end.
